SIMPLE_WP_MODEL = None
ADV_WP_MODEL = None


def state_win_probability(frame, model):
    """Predicts the win probability of a given frame.

    Args:
        frame (dict): Dict output of a frame generated from the DemoParser class

    Returns:
        A dictionary containing the CT and T round win probabilities
    """
    raise NotImplementedError


def round_win_probability(ct_score, t_score, map):
    """Estimates of game win probability using information from the HLTV win matrix for a given map and score.

    Args:
        ct_score (dict): CT Score
        t_score (dict): T Score
        map (dict): Map the demo is from

    Returns:
        A dictionary containing the CT game win, T game win and Draw probabilities
    """
    return NotImplementedError



""" Util functions for csgo package
"""

import json
import numpy as np
import re
import subprocess


class AutoVivification(dict):
    """Implementation of perl's autovivification feature. Stolen from https://stackoverflow.com/questions/651794/whats-the-best-way-to-initialize-a-dict-of-dicts-in-python"""

    def __getitem__(self, item):
        try:
            return dict.__getitem__(self, item)
        except KeyError:
            value = self[item] = type(self)()
            return value


def check_go_version():
    """Function to check the Golang version of the current machine, returns True if greater than 1.14.0"""
    try:
        proc = subprocess.Popen(["go", "version"], stdout=subprocess.PIPE)
        parsed_resp = proc.stdout.read().splitlines()
        if len(parsed_resp) != 1:
            raise ValueError("Error finding Go version")
        else:
            go_version_text = parsed_resp[0].decode("utf-8")
            go_version = re.findall(r"\d\.\d+", go_version_text)
            if int(go_version[0].replace(".", "")) >= 117:
                return True
            else:
                return False
    except Exception as e:
        print(e)
        return False


def is_in_range(value, min, max):
    if value >= min and value <= max:
        return True
    else:
        return False


def transform_csv_to_json(sampleCsv):
    """From Adi. Used to transform a nav file CSV to JSON."""
    finalDic = {}
    for curMap in sampleCsv["mapName"].unique():
        mapDic = {}
        for i in sampleCsv[sampleCsv["mapName"] == curMap].index:
            curTile = sampleCsv.iloc[i]
            curDic = {}
            for curFeature in sampleCsv.columns:
                if curFeature not in ["mapName", "areaId"]:
                    curDic[curFeature] = curTile[curFeature]
            mapDic[curTile["areaId"]] = curDic
        finalDic[curMap] = mapDic
    return finalDic



""" Data cleaning functions
"""

import difflib
import numpy as np
import pandas as pd
import textdistance


def associate_entities(game_names=[], entity_names=[], metric="lcss"):
    """A function to return a dict of associated entities. Accepts

    Args:
        game_names (list): A list of names generated by the demofile
        entity_names (list): A list of names
        metric (string): A string indicating distance metric, one of lcss, hamming, levenshtein, jaro, difflib

    Returns:
        A dictionary where the keys are entries in game_names, values are the matched entity names.
    """
    if metric.lower() == "lcss":
        dist_metric = textdistance.lcsseq.distance
    elif metric.lower() == "hamming":
        dist_metric = textdistance.hamming.distance
    elif metric.lower() == "levenshtein":
        dist_metric = textdistance.levenshtein.distance
    elif metric.lower() == "jaro":
        dist_metric = textdistance.jaro.distance
    elif metric.lower() == "difflib":
        entities = {}
        for gn in game_names:
            if gn is not None and gn is not np.nan:
                closest_name = difflib.get_close_matches(
                    gn, entity_names, n=1, cutoff=0.0
                )
                if len(closest_name) > 0:
                    entities[gn] = closest_name[0]
                else:
                    entities[gn] = None
        entities[None] = None
        return entities
    else:
        raise ValueError(
            "Metric can only be lcss, hamming, levenshtein, jaro or difflib"
        )
    entities = {}
    for gn in game_names:
        if gn is not None and gn is not np.nan and gn != "":
            name_distances = []
            names = []
            if len(entity_names) > 0:
                for p in entity_names:
                    name_distances.append(dist_metric(gn.lower(), p.lower()))
                    names.append(p)
                entities[gn] = names[np.argmin(name_distances)]
                popped_name = entity_names.pop(np.argmin(name_distances))
        if gn == "":
            entities[gn] = None
    entities[None] = None
    return entities


def replace_entities(df, col_name, entity_dict):
    """A function to replace values in a Pandas df column given an entity dict, as created in associate_entities()

    Args:
        df (DataFrame)     : A Pandas DataFrame
        col_name (string)  : A column in the Pandas DataFrame
        entity_dict (dict) : A dictionary as created in the associate_entities() function

    Returns:
        A dataframe with replaced names.
    """
    if col_name not in df.columns:
        raise ValueError("Column does not exist!")
    df[col_name].replace(entity_dict, inplace=True)
    return df



def generate_game_state(frame, state_type="vector"):
    """Returns a frame into a game state form (one of vector, graph or set)

    Args:
        frame (dict)        : Dict output of a frame generated from the DemoParser class
        state_type (string) : One of "vector", "graph" or "set"

    Returns:
        A dict with keys "T", "CT" and "Global" containing the state representation
    """
    if type(frame) is not dict:
        raise ValueError(
            "Frame input must be a dict from the DemoParser.parse() output"
        )

    if state_type not in ["vector", "graph", "set"]:
        raise ValueError("Supported state types are vector, graph and set.")

    state = None

    if state_type == "vector":
        state = generate_vector_state(frame)

    if state_type == "graph":
        state = generate_graph_state(frame)

    if state_type == "set":
        state = generate_set_state(frame)

    return state


def generate_team_vector_state(frame_side):
    """Returns a team's game state as a vector. The included features are the total players remaining, full hp players remaining, hp, armor, helmets, equipment value and defuse kits.

    Args:
        frame_side (dict) : Dict output of a frame side generated from the DemoParser class

    Returns:
        A list with numeric elements
    """
    team_state = {}
    team_state["eqVal"] = 0
    team_state["playersRemaining"] = 0
    team_state["fullHpPlayersRemaining"] = 0
    team_state["hp"] = 0
    team_state["armor"] = 0
    team_state["helmets"] = 0
    team_state["defuseKits"] = 0
    for player in frame_side["players"]:
        if player["isAlive"]:
            team_state["eqVal"] += player["equipmentValue"]
            team_state["playersRemaining"] += 1
            team_state["hp"] += player["hp"]
            team_state["armor"] += player["armor"]
            if player["hasHelmet"]:
                team_state["helmets"] += 1
            if player["hasDefuse"]:
                team_state["defuseKits"] += 1
            if player["hp"] == 100:
                team_state["fullHpPlayersRemaining"] += 1
    return team_state


def generate_world_vector_state(frame):
    """Generate's the world state as a vector

    Args:
        frame (dict) : Dict output of a frame generated from the DemoParser class

    Returns:
        A list with numeric elements
    """
    world_state = {}
    world_state["bombPlanted"] = 0
    if frame["bombPlanted"]:
        world_state["bombPlanted"] = 1
    world_state["secondsRemainingInPhase"] = frame["seconds"]
    world_state["bombsite"] = frame["bombsite"]
    return world_state


def generate_vector_state(frame):
    """Returns a game state as a vector. The vector includes the following information:

    For each team:
        - players remaining
        - full hp players remaining
        - total hp
        - total armor
        - total helmets
        - total equipment value
        - total defuse kits

    For the world state:
        - seconds since the beginning of the game phase (bomb planted/not)
        - bomb plant
        - bomb site where bomb is planted

    Args:
        frame (dict) : Dict output of a frame generated from the DemoParser class

    Returns:
        A dict with keys "T", "CT" and "Global", where each entry is a vector. Global vector is CT + T concatenated
    """
    state = {}
    state["ct"] = generate_team_vector_state(frame["ct"])
    state["t"] = generate_team_vector_state(frame["t"])
    state["global"] = generate_world_vector_state(frame)
    return state


def generate_graph_state(frame):
    """Returns a game state as a graph

    Args:
        frame (dict) : Dict output of a frame generated from the DemoParser class

    Returns:
        A dict with keys "T", "CT" and "Global", where each entry is a vector. Global vector is CT + T concatenated
    """
    return {"ct": [], "t": [], "global": []}


def generate_set_state(frame):
    """Returns a game state as a set

    Args:
        frame (dict) : Dict output of a frame generated from the DemoParser class

    Returns:
        A dict with keys "T", "CT" and "Global", where each entry is a vector. Global vector is CT + T concatenated
    """
    return {"ct": [], "t": [], "global": []}



from multiprocessing.sharedctypes import Value
import networkx as nx
import numpy as np

from awpy.data import NAV, NAV_GRAPHS
from scipy.spatial import distance


def point_in_area(map_name, area_id, point):
    """Returns if the point is within a nav area for a map.

    Args:
        map_name (string): Map to search
        area_id (int): Area ID as an integer
        point (list): Point as a list [x,y,z]

    Returns:
        True if area contains the point, false if not
    """
    if map_name not in NAV.keys():
        raise ValueError("Map not found.")
    if area_id not in NAV[map_name].keys():
        raise ValueError("Area ID not found.")
    if len(point) != 3:
        raise ValueError("Point must be a list [X,Y,Z]")
    contains_x = (
        min(NAV[map_name][area_id]["northWestX"], NAV[map_name][area_id]["southEastX"])
        < point[0]
        < max(
            NAV[map_name][area_id]["northWestX"], NAV[map_name][area_id]["southEastX"]
        )
    )
    contains_y = (
        min(NAV[map_name][area_id]["northWestY"], NAV[map_name][area_id]["southEastY"])
        < point[1]
        < max(
            NAV[map_name][area_id]["northWestY"], NAV[map_name][area_id]["southEastY"]
        )
    )
    if contains_x and contains_y:
        return True
    else:
        return False


def find_closest_area(map_name, point):
    """Finds the closest area in the nav mesh. Searches through all the areas by comparing point to area centerpoint.

    Args:
        map_name (string): Map to search
        point (list): Point as a list [x,y,z]

    Returns:
        A dict containing info on the closest area
    """
    if map_name not in NAV.keys():
        raise ValueError("Map not found.")
    if len(point) != 3:
        raise ValueError("Point must be a list [X,Y,Z]")
    closest_area = {"mapName": map_name, "areaId": None, "distance": 999999}
    for area in NAV[map_name].keys():
        avg_x = (
            NAV[map_name][area]["northWestX"] + NAV[map_name][area]["southEastX"]
        ) / 2
        avg_y = (
            NAV[map_name][area]["northWestY"] + NAV[map_name][area]["southEastY"]
        ) / 2
        avg_z = (
            NAV[map_name][area]["northWestZ"] + NAV[map_name][area]["southEastZ"]
        ) / 2
        dist = np.sqrt(
            (point[0] - avg_x) ** 2 + (point[1] - avg_y) ** 2 + (point[2] - avg_z) ** 2
        )
        if dist < closest_area["distance"]:
            closest_area["areaId"] = area
            closest_area["distance"] = dist
    return closest_area


def area_distance(map_name, area_a, area_b, dist_type="graph"):
    """Returns the distance between two areas. Dist type an be graph or geodesic.

    Args:
        map_name (string): Map to search
        area_a (int): Area id
        area_b (int): Area id
        dist_type (string): String indicating the type of distance to use (graph or geodesic)

    Returns:
        A dict containing info on the path between two areas.
    """
    if map_name not in NAV.keys():
        raise ValueError("Map not found.")
    if (area_a not in NAV[map_name].keys()) or (area_b not in NAV[map_name].keys()):
        raise ValueError("Area ID not found.")
    if dist_type not in ["graph", "geodesic"]:
        raise ValueError("dist_type can only be graph or geodesic")
    G = NAV_GRAPHS[map_name]
    distance_obj = {"distanceType": dist_type, "distance": None, "areas": []}
    if dist_type == "graph":
        discovered_path = nx.shortest_path(G, area_a, area_b)
        distance_obj["distance"] = len(discovered_path) - 1
        distance_obj["areas"] = discovered_path
        return distance_obj
    if dist_type == "geodesic":

        def dist(a, b):
            return G.nodes()[a]["size"] + G.nodes()[b]["size"]

        geodesic_path = nx.astar_path(G, area_a, area_b, heuristic=dist)
        geodesic_cost = 0
        for i, area in enumerate(geodesic_path):
            if i > 0:
                geodesic_cost += G.nodes()[area]["size"]
        distance_obj["distance"] = geodesic_cost
        distance_obj["areas"] = geodesic_path
        return distance_obj


def point_distance(map_name, point_a, point_b, dist_type="graph"):
    """Returns the distance between two points.

    Args:
        map_name (string): Map to search
        point_a (list): Point as a list (x,y,z)
        point_b (list): Point as a list (x,y,z)
        dist_type (string): String indicating the type of distance to use. Can be graph, geodesic, euclidena, manhattan, canberra or cosine.

    Returns:
        A dict containing info on the distance between two points.
    """
    distance_obj = {"distanceType": dist_type, "distance": None, "areas": []}
    if dist_type == "graph":
        if map_name not in NAV.keys():
            raise ValueError("Map not found.")
        if len(point_a) != 3 or len(point_b) != 3:
            raise ValueError(
                "When using graph or geodesic distance, point must be X/Y/Z"
            )
        area_a = find_closest_area(map_name, point_a)["areaId"]
        area_b = find_closest_area(map_name, point_b)["areaId"]
        return area_distance(map_name, area_a, area_b, dist_type=dist_type)
    elif dist_type == "geodesic":
        if map_name not in NAV.keys():
            raise ValueError("Map not found.")
        if len(point_a) != 3 or len(point_b) != 3:
            raise ValueError(
                "When using graph or geodesic distance, point must be X/Y/Z"
            )
        area_a = find_closest_area(map_name, point_a)["areaId"]
        area_b = find_closest_area(map_name, point_b)["areaId"]
        return area_distance(map_name, area_a, area_b, dist_type=dist_type)
    elif dist_type == "euclidean":
        distance_obj["distance"] = distance.euclidean(point_a, point_b)
        return distance_obj
    elif dist_type == "manhattan":
        distance_obj["distance"] = distance.cityblock(point_a, point_b)
        return distance_obj
    elif dist_type == "canberra":
        distance_obj["distance"] = distance.canberra(point_a, point_b)
        return distance_obj
    elif dist_type == "cosine":
        distance_obj["distance"] = distance.cosine(point_a, point_b)
        return distance_obj


def generate_position_token(map_name, frame):
    """Generates the position token for a game frame.

    Args:
        map_name (string): Map to search
        frame (dict): A game frame

    Returns:
        A dict containing the T token, CT token and combined token (T + CT concatenated)
    """
    if map_name not in NAV.keys():
        raise ValueError("Map not found.")
    if (len(frame["ct"]["players"]) == 0) or (len(frame["t"]["players"]) == 0):
        raise ValueError("CT or T players has length of 0")
    # Create map area list
    map_area_names = []
    for area_id in NAV[map_name]:
        if NAV[map_name][area_id]["areaName"] not in map_area_names:
            map_area_names.append(NAV[map_name][area_id]["areaName"])
    map_area_names.sort()
    # Create token
    ct_token = np.zeros(len(map_area_names), dtype=np.int8)
    for player in frame["ct"]["players"]:
        if player["isAlive"]:
            closest_area = find_closest_area(
                map_name, [player["x"], player["y"], player["z"]]
            )
            ct_token[
                map_area_names.index(NAV[map_name][closest_area["areaId"]]["areaName"])
            ] += 1
    t_token = np.zeros(len(map_area_names), dtype=np.int8)
    for player in frame["t"]["players"]:
        if player["isAlive"]:
            closest_area = find_closest_area(
                map_name, [player["x"], player["y"], player["z"]]
            )
            t_token[
                map_area_names.index(NAV[map_name][closest_area["areaId"]]["areaName"])
            ] += 1
    # Create payload
    token = {}
    token["tToken"] = (
        str(t_token).replace("'", "").replace("[", "").replace("]", "").replace(" ", "")
    )
    token["ctToken"] = (
        str(ct_token)
        .replace("'", "")
        .replace("[", "")
        .replace("]", "")
        .replace(" ", "")
    )
    token["token"] = token["ctToken"] + token["tToken"]
    return token



import pandas as pd

# accuracy
# kast
# adr
# kill stats
# flash stats
# econ stats
def player_stats(game_rounds, return_type="json"):
    player_statistics = {}
    for r in game_rounds:
        # Add players
        ct_side = r["ctSide"]
        t_side = r["tSide"]
        for p in ct_side["players"]:
            if p["steamID"] not in player_statistics:
                player_statistics[p["steamID"]] = {
                    "steamID": p["steamID"],
                    "playerName": p["playerName"],
                    "teamName": ct_side["teamName"],
                    "totalRounds": 0,
                    "kills": 0,
                    "deaths": 0,
                    "kdr": 0,
                    "assists": 0,
                    "tradeKills": 0,
                    "teamKills": 0,
                    "suicides": 0,
                    "flashAssists": 0,
                    "totalDamageGiven": 0,
                    "totalDamageTaken": 0,
                    "totalTeamDamageGiven": 0,
                    "adr": 0,
                    "totalShots": 0,
                    "shotsHit": 0,
                    "accuracy": 0,
                    "rating": 0,
                    "kast": 0,
                    "hs": 0,
                    "hsPercent": 0,
                    "firstKills": 0,
                    "firstDeaths": 0,
                    "utilityDamage": 0,
                    "smokesThrown": 0,
                    "flashesThrown": 0,
                    "heThrown": 0,
                    "fireThrown": 0,
                    "enemiesFlashed": 0,
                    "teammatesFlashed": 0,
                    "blindTime": 0,
                    "plants": 0,
                    "defuses": 0,
                }
                player_statistics[p["steamID"]]["totalRounds"] += 1
            else:
                player_statistics[p["steamID"]]["totalRounds"] += 1
        for p in t_side["players"]:
            if p["steamID"] not in player_statistics:
                player_statistics[p["steamID"]] = {
                    "steamID": p["steamID"],
                    "playerName": p["playerName"],
                    "teamName": t_side["teamName"],
                    "totalRounds": 0,
                    "kills": 0,
                    "deaths": 0,
                    "kdr": 0,
                    "assists": 0,
                    "tradeKills": 0,
                    "teamKills": 0,
                    "suicides": 0,
                    "flashAssists": 0,
                    "totalDamageGiven": 0,
                    "totalDamageTaken": 0,
                    "totalTeamDamageGiven": 0,
                    "adr": 0,
                    "totalShots": 0,
                    "shotsHit": 0,
                    "accuracy": 0,
                    "rating": 0,
                    "kast": 0,
                    "hs": 0,
                    "hsPercent": 0,
                    "firstKills": 0,
                    "firstDeaths": 0,
                    "utilityDamage": 0,
                    "smokesThrown": 0,
                    "flashesThrown": 0,
                    "heThrown": 0,
                    "fireThrown": 0,
                    "enemiesFlashed": 0,
                    "teammatesFlashed": 0,
                    "blindTime": 0,
                    "plants": 0,
                    "defuses": 0,
                }
                player_statistics[p["steamID"]]["totalRounds"] += 1
            else:
                player_statistics[p["steamID"]]["totalRounds"] += 1
        kast = {}
        for p in t_side["players"]:
            kast[p["steamID"]] = {}
            kast[p["steamID"]]["k"] = False
            kast[p["steamID"]]["a"] = False
            kast[p["steamID"]]["s"] = True
            kast[p["steamID"]]["t"] = False
        for p in ct_side["players"]:
            kast[p["steamID"]] = {}
            kast[p["steamID"]]["k"] = False
            kast[p["steamID"]]["a"] = False
            kast[p["steamID"]]["s"] = True
            kast[p["steamID"]]["t"] = False
        # Calculate kills
        for i, k in enumerate(r["kills"]):
            if (
                k["attackerSteamID"]
                and k["attackerSteamID"] in player_statistics.keys()
            ):
                player_statistics[k["attackerSteamID"]]["kills"] += 1
                kast[k["attackerSteamID"]]["k"] = True
            if k["victimSteamID"] and k["victimSteamID"] in player_statistics.keys():
                player_statistics[k["victimSteamID"]]["deaths"] += 1
                kast[k["victimSteamID"]]["s"] = False
            if (
                k["assisterSteamID"]
                and k["assisterSteamID"] in player_statistics.keys()
            ):
                player_statistics[k["assisterSteamID"]]["assists"] += 1
                kast[k["assisterSteamID"]]["a"] = True
            if (
                k["flashThrowerSteamID"]
                and k["flashThrowerSteamID"] in player_statistics.keys()
            ):
                player_statistics[k["flashThrowerSteamID"]]["flashAssists"] += 1
                kast[k["flashThrowerSteamID"]]["a"] = True
            if (
                k["isTrade"]
                and k["attackerSteamID"]
                and k["attackerSteamID"] in player_statistics.keys()
            ):
                player_statistics[k["attackerSteamID"]]["tradeKills"] += 1
                kast[r["kills"][i - 1]["victimSteamID"]]["t"] = True
            if (
                k["isFirstKill"]
                and k["attackerSteamID"]
                and k["attackerSteamID"] in player_statistics.keys()
            ):
                player_statistics[k["attackerSteamID"]]["firstKills"] += 1
                player_statistics[k["victimSteamID"]]["firstDeaths"] += 1
            if (
                k["isTeamkill"]
                and k["attackerSteamID"]
                and k["attackerSteamID"] in player_statistics.keys()
            ):
                player_statistics[k["attackerSteamID"]]["teamKills"] += 1
            if (
                k["isHeadshot"]
                and k["attackerSteamID"]
                and k["attackerSteamID"] in player_statistics.keys()
            ):
                player_statistics[k["attackerSteamID"]]["hs"] += 1
            if k["isSuicide"] and k["victimSteamID"] in player_statistics.keys():
                player_statistics[k["victimSteamID"]]["suicides"] += 1
        for d in r["damages"]:
            if (
                d["attackerSteamID"]
                and not d["isFriendlyFire"]
                and d["attackerSteamID"] in player_statistics.keys()
            ):
                player_statistics[d["attackerSteamID"]]["totalDamageGiven"] += d[
                    "hpDamageTaken"
                ]
            if d["victimSteamID"] and d["victimSteamID"] in player_statistics.keys():
                player_statistics[d["victimSteamID"]]["totalDamageTaken"] += d[
                    "hpDamageTaken"
                ]
            if d["isFriendlyFire"] and d["attackerSteamID"] in player_statistics.keys():
                player_statistics[d["attackerSteamID"]]["totalTeamDamageGiven"] += d[
                    "hpDamageTaken"
                ]
            if (
                d["weaponClass"] not in ["Unknown", "Grenade", "Equipment"]
                and d["attackerSteamID"] in player_statistics.keys()
            ):
                player_statistics[d["attackerSteamID"]]["shotsHit"] += 1
            if (
                d["weaponClass"] == "Grenade"
                and d["attackerSteamID"] in player_statistics.keys()
            ):
                player_statistics[d["attackerSteamID"]]["utilityDamage"] += d[
                    "hpDamageTaken"
                ]
        for w in r["weaponFires"]:
            if w["playerSteamID"] in player_statistics.keys():
                player_statistics[w["playerSteamID"]]["totalShots"] += 1
        for f in r["flashes"]:
            if (
                f["attackerSteamID"]
                and f["attackerSteamID"] in player_statistics.keys()
            ):
                if f["attackerSide"] == f["playerSide"]:
                    player_statistics[f["attackerSteamID"]]["teammatesFlashed"] += 1
                else:
                    player_statistics[f["attackerSteamID"]]["enemiesFlashed"] += 1
                    player_statistics[f["attackerSteamID"]]["blindTime"] += f[
                        "flashDuration"
                    ]
        for g in r["grenades"]:
            if g["throwerSteamID"] and g["throwerSteamID"] in player_statistics.keys():
                if g["grenadeType"] == "Smoke Grenade":
                    player_statistics[g["throwerSteamID"]]["smokesThrown"] += 1
                if g["grenadeType"] == "Flashbang":
                    player_statistics[g["throwerSteamID"]]["flashesThrown"] += 1
                if g["grenadeType"] == "HE Grenade":
                    player_statistics[g["throwerSteamID"]]["heThrown"] += 1
                if g["grenadeType"] in ["Incendiary Grenade", "Molotov"]:
                    player_statistics[g["throwerSteamID"]]["fireThrown"] += 1
        for b in r["bombEvents"]:
            if b["playerSteamID"] and b["playerSteamID"] in player_statistics.keys():
                if b["bombAction"] == "plant":
                    player_statistics[b["playerSteamID"]]["plants"] += 1
                if b["bombAction"] == "defuse":
                    player_statistics[b["playerSteamID"]]["defuses"] += 1
        for player in kast:
            all_true = False
            for component in kast[player]:
                if kast[player][component]:
                    all_true = True
            if all_true:
                player_statistics[player]["kast"] += 1
    for player in kast:
        player_statistics[player]["kast"] = round(
            100
            * player_statistics[player]["kast"]
            / player_statistics[player]["totalRounds"],
            1,
        )
    for player in player_statistics:
        player_statistics[player]["blindTime"] = round(
            player_statistics[player]["blindTime"], 2
        )
    for player in player_statistics:
        player_statistics[player]["kdr"] = round(
            player_statistics[player]["kills"] / player_statistics[player]["deaths"]
            if player_statistics[player]["deaths"] != 0
            else player_statistics[player]["kills"],
            2,
        )
    for player in player_statistics:
        player_statistics[player]["adr"] = round(
            player_statistics[player]["totalDamageGiven"]
            / player_statistics[player]["totalRounds"],
            1,
        )
    for player in player_statistics:
        player_statistics[player]["accuracy"] = round(
            player_statistics[player]["shotsHit"]
            / player_statistics[player]["totalShots"],
            2,
        )
    for player in player_statistics:
        player_statistics[player]["hsPercent"] = round(
            player_statistics[player]["hs"] / player_statistics[player]["kills"]
            if player_statistics[player]["kills"] != 0
            else 0,
            2,
        )
    for player in player_statistics:
        impact = (
            2.13
            * (
                player_statistics[player]["kills"]
                / player_statistics[player]["totalRounds"]
            )
            + 0.42
            * (
                player_statistics[player]["assists"]
                / player_statistics[player]["totalRounds"]
            )
            - 0.41
        )
        player_statistics[player]["rating"] = (
            0.0073 * player_statistics[player]["kast"]
            + 0.3591
            * (
                player_statistics[player]["kills"]
                / player_statistics[player]["totalRounds"]
            )
            - 0.5329
            * (
                player_statistics[player]["deaths"]
                / player_statistics[player]["totalRounds"]
            )
            + 0.2372 * (impact)
            + 0.0032 * (player_statistics[player]["adr"])
            + 0.1587
        )
        player_statistics[player]["rating"] = round(
            player_statistics[player]["rating"], 2
        )
    if return_type == "df":
        return (
            pd.DataFrame()
            .from_dict(player_statistics, orient="index")
            .reset_index(drop=True)
        )
    else:
        return player_statistics


# def accuracy(data):
#     """_summary_

#     Args:
#         data (_type_): _description_
#     """
#     accuracy_data = {}
#     for r in data:
#         if r["weaponFires"]:
#             for w in r["weaponFires"]:
#                 if w["playerSteamID"] not in accuracy_data:
#                     accuracy_data[w["playerSteamID"]] = {}
#                     accuracy_data[w["playerSteamID"]]["playerName"] = w["playerName"]
#                     accuracy_data[w["playerSteamID"]]["shotsFired"] = 1
#                     accuracy_data[w["playerSteamID"]]["shotsHit"] = 0
#                 else:
#                     accuracy_data[w["playerSteamID"]]["shotsFired"] += 1
#             for d in r["damages"]:
#                 if (
#                     (d["weaponClass"] != "Grenades")
#                     and (d["weaponClass"] != "Equipment")
#                     and (d["weaponClass"] != "Unknown")
#                 ):
#                     accuracy_data[w["playerSteamID"]]["shotsHit"] += 1
#     return accuracy_data


# def kast(
#     data,
#     flash_assists=True,
#     return_type="json",
# ):
#     """Creates KAST% for a given round/kill dataframe combination.

#     Args:
#         data (dict): Default
#         flash_assists (bool, optional): _description_. Defaults to True.
#         return_type (str, optional): _description_. Defaults to "json".
#     """
#     kast_data = {}
#     for r in data["gameRounds"]:
#         if r["kills"]:
#             for k in r["kills"]:
#                 if k["attackerName"] is not None:
#                     # Count kills
#                     if k["attackerName"] in kast_data:
#                         kast_data[k["attackerName"]]["kills"] += 1
#                     else:
#                         kast_data[k["attackerName"]]["kills"] = 1
#                     # Count assists
#     pass


# def extract_num_filters(
#     filters: Dict[str, Union[List[bool], List[str]]], key: str
# ) -> Tuple[List[str], List[float]]:
#     """Extracts the numeric column filters from a key-value pair in a dictionary.

#     Extracts logical operators and numeric values from the column filters in the
#     the value of a specified key in the dictionary represented by filters.

#     Args:
#         filters: A dictionary where the keys are the columns of the dataframe
#             represented by df to filter the data by and the values are lists
#             that contain the column filters.
#         key: The key in the dictionary represented by filters for which the
#             numeric column filters should be extracted.

#     Returns:
#         A tuple of two lists where the first list contains all of the logical
#         operators in the column filters in the value of a specified key and the
#         second list contains all of the numeric values in the column filters in
#         the value of a specified key. For example:
#         ([">", "<"], [15.0, 25.0])

#     Raises:
#         ValueError: The value of the specified key contains a column filter of
#             type boolean.
#         Exception: The column filters in the value of the specified key contain
#             an invalid logical operator.
#         Exception: The column filters in the value of the specified key contain
#             an invalid numeric value.
#     """
#     sign_list = []
#     val_list = []
#     for index in filters[key]:
#         if not isinstance(index, str):
#             raise ValueError(
#                 f'Filter(s) for column "{key}" must be of type ' f"string."
#             )
#         i = 0
#         sign = ""
#         while i < len(index) and not index[i].isdecimal():
#             sign += index[i]
#             end_index = i
#             i += 1
#         if sign not in ("==", "!=", "<=", ">=", "<", ">"):
#             raise Exception(
#                 f'Invalid logical operator in filters for "{key}"' f" column."
#             )
#         sign_list.append(sign)
#         try:
#             val_list.append(float(index[end_index + 1 :]))
#         except ValueError as ve:
#             raise Exception(
#                 f'Invalid numerical value in filters for "{key}" ' f"column."
#             ) from ve
#     return sign_list, val_list


# def check_filters(df: pd.DataFrame, filters: Dict[str, Union[List[bool], List[str]]]):
#     """Checks if the column filters are valid.

#     Iterates through each key-value pair in the dictionary represented by
#     filters and checks if the column filters for columns of type boolean are
#     of type boolean and the column filters for columns of type string are of
#     type string. Calls the function extract_numeric_filters to check if the
#     column filters for columns of type float and integer are valid.

#     Args:
#         df: A dataframe to be filtered.
#         filters: A dictionary where the keys are the columns of the dataframe
#             represented by df to filter the data by and the values are lists
#             that contain the column filters.

#     Raises:
#         ValueError: The value of a key corresponding to a column of type boolean
#             contains a column filter of a different type.
#         ValueError: The value of a key corresponding to a column of type string
#             contains a column filter of a different type.
#     """
#     for key in filters:
#         if df.dtypes[key] == "bool":
#             for index in filters[key]:
#                 if not isinstance(index, bool):
#                     raise ValueError(
#                         f'Filter(s) for column "{key}" must be ' f"of type boolean"
#                     )
#         elif df.dtypes[key] == "O":
#             for index in filters[key]:
#                 if not isinstance(index, str):
#                     raise ValueError(
#                         f'Filter(s) for column "{key}" must be ' f"of type string"
#                     )
#         else:
#             extract_num_filters(filters, key)


# def num_filter_df(df: pd.DataFrame, col: str, sign: str, val: float) -> pd.DataFrame:
#     """Filters numeric data in a dataframe.

#     Args:
#         df: A dataframe to be filtered.
#         col: A column of the dataframe represented by df to filter by.
#         sign: A logical operator representing the operation to filter the
#             dataframe represented by df by.
#         val: A numeric value to filter the dataframe represented by df by.

#     Returns:
#         A filtered dataframe.
#     """
#     ops = {
#         "==": operator.eq(df[col], val),
#         "!=": operator.ne(df[col], val),
#         "<=": operator.le(df[col], val),
#         ">=": operator.ge(df[col], val),
#         "<": operator.lt(df[col], val),
#         ">": operator.gt(df[col], val),
#     }
#     filtered_df = df.loc[ops[sign]]
#     return filtered_df


# def filter_df(
#     df: pd.DataFrame, filters: Dict[str, Union[List[bool], List[str]]]
# ) -> pd.DataFrame:
#     """Filters data in a dataframe.

#     Filters a dataframe for columns of type boolean and string and calls the
#     function num_filter_df to filter a dataframe for columns of type float
#     and integer.

#     Args:
#         df: A dataframe to be filtered.
#         filters: A dictionary where the keys are the columns of the dataframe
#             represented by df to filter the data by and the values are lists
#             that contain the column filters.

#     Returns:
#         A filtered dataframe.
#     """
#     df_copy = df.copy()
#     check_filters(df_copy, filters)
#     for key in filters:
#         if df_copy.dtypes[key] == "bool" or df_copy.dtypes[key] == "O":
#             df_copy = df_copy.loc[df_copy[key].isin(filters[key])]
#         else:
#             i = 0
#             for _ in extract_num_filters(filters, key)[0]:
#                 val = extract_num_filters(filters, key)[1][i]
#                 df_copy = num_filter_df(
#                     df_copy, key, extract_num_filters(filters, key)[0][i], val
#                 )
#                 i += 1
#     return df_copy


# def calc_stats(
#     df: pd.DataFrame,
#     filters: Dict[str, Union[List[bool], List[str]]],
#     col_to_groupby: List[str],
#     col_to_agg: List[str],
#     agg: List[List[str]],
#     col_names: List[str],
# ) -> pd.DataFrame:
#     """Calculates statistics for a dataframe.

#     Calls the function filter_df to filter a data in dataframe and then performs
#     groupby and aggregation operations on the dataframe and renames the columns.

#     Args:
#         df: A dataframe from which statistics should be calculated.
#         filters: A dictionary where the keys are the columns of the dataframe
#             represented by df to filter the data by and the values are lists
#             that contain the column filters.
#         col_to_groupby: Columns of the dataframe represented by df to perform
#             the groupby operation on.
#         col_to_agg: Columns of the dataframe represented by df to perform
#             aggregation operations on.
#         agg: The aggregation operations to be applied to the columns in the list
#             represented by col_to_agg.
#         col_names: Column names for the returned dataframe.

#     Returns:
#         A dataframe with the specified statistics.
#     """
#     df_copy = filter_df(df, filters)
#     agg_dict = dict(zip(col_to_agg, agg))
#     if col_to_agg:
#         df_copy = df_copy.groupby(col_to_groupby).agg(agg_dict).reset_index()
#     df_copy.columns = col_names
#     return df_copy


# def accuracy(
#     damage_data: pd.DataFrame,
#     weapon_fire_data: pd.DataFrame,
#     team: bool = False,
#     damage_filters: Dict[str, Union[List[bool], List[str]]] = {},
#     weapon_fire_filters: Dict[str, Union[List[bool], List[str]]] = {},
# ) -> pd.DataFrame:
#     """Returns a dataframe with accuracy statistics.

#     Args:
#         damage_data: A dataframe with damage data.
#         weapon_fire_data: A dataframe with weapon fire data.
#         team: A boolean specifying wheter to calculate statistics for each team
#             or for each player. The default is to calculate statistics for
#             each player.
#         damage_filters: A dictionary where the keys are the columns of the
#             dataframe represented by damage_data to filter the damage data by
#             and the values are lists that contain the column filters.
#         weapon_fire_filters: A dictionary where the keys are the columns of the
#             dataframe represented by weapon_fires to filter the weapon fire data
#             by and the values are lists that contain the column filters.

#     Returns:
#         A dataframe with the total shots, strafe shots, hits, accuracy (hits/total number of shots), headshot accuracy (head hits/total number of shots)
#     """
#     stats = ["playerName", "attackerName", "Player"]
#     if team:
#         stats = ["playerTeam", "attackerTeam", "Team"]
#     weapon_fires = calc_stats(
#         weapon_fire_data,
#         weapon_fire_filters,
#         [stats[0]],
#         [stats[0]],
#         [["size"]],
#         [stats[2], "Weapon Fires"],
#     )
#     strafe_fires = calc_stats(
#         weapon_fire_data.loc[weapon_fire_data["playerStrafe"] == True],
#         weapon_fire_filters,
#         [stats[0]],
#         [stats[0]],
#         [["size"]],
#         [stats[2], "Strafe Fires"],
#     )
#     hits = calc_stats(
#         damage_data.loc[damage_data["attackerTeam"] != damage_data["victimTeam"]],
#         damage_filters,
#         [stats[1]],
#         [stats[1]],
#         [["size"]],
#         [stats[2], "Hits"],
#     )
#     headshots = calc_stats(
#         damage_data.loc[
#             (damage_data["attackerTeam"] != damage_data["victimTeam"])
#             & (damage_data["hitGroup"] == "Head")
#         ],
#         damage_filters,
#         [stats[1]],
#         [stats[1]],
#         [["size"]],
#         [stats[2], "Headshots"],
#     )
#     acc = weapon_fires.merge(strafe_fires, how="outer")  # .fillna(0)
#     acc = acc.merge(hits, how="outer")  # .fillna(0)
#     acc = acc.merge(headshots, how="outer")  # .fillna(0)
#     acc["Strafe%"] = acc["Strafe Fires"] / acc["Weapon Fires"]
#     acc["ACC%"] = acc["Hits"] / acc["Weapon Fires"]
#     acc["HS ACC%"] = acc["Headshots"] / acc["Weapon Fires"]
#     acc = acc[[stats[2], "Weapon Fires", "Strafe%", "ACC%", "HS ACC%"]]
#     acc.sort_values(by="ACC%", ascending=False, inplace=True)
#     acc.reset_index(drop=True, inplace=True)
#     return acc


# def kast(
#     kill_data: pd.DataFrame,
#     kast_string: str = "KAST",
#     flash_assists: bool = True,
#     kill_filters: Dict[str, Union[List[bool], List[str]]] = {},
#     death_filters: Dict[str, Union[List[bool], List[str]]] = {},
# ) -> pd.DataFrame:
#     """Returns a dataframe with KAST statistics.

#     Args:
#         kill_data: A dataframe with kill data.
#         kast_string: A string specifying which combination of KAST statistics
#             to use.
#         flash_assists: A boolean specifying if flash assists are to be
#             counted as assists or not.
#         kill_filters: A dictionary where the keys are the columns of the
#             dataframe represented by kill_data to filter the kill data by and
#             the values are lists that contain the column filters.
#         death_filters: A dictionary where the keys are the columns of the
#             dataframe represented by kill_data to filter the death data by and
#             the values are lists that contain the column filters.

#     Returns:
#         A dataframe containing the percentage of a player's rounds with at least one of (K)ill, (A)ssist, (S)urvival and (T)rades.
#     """
#     columns = ["Player", f"{kast_string.upper()}%"]
#     kast_counts = {}
#     kast_rounds = {}
#     for stat in kast_string.upper():
#         columns.append(stat)
#     kill_data = kill_data[~kill_data["attackerName"].isna()]
#     killers = calc_stats(
#         kill_data.loc[
#             (kill_data["attackerTeam"] != kill_data["victimTeam"])
#             & (kill_data["attackerTeam"].notnull())
#         ],
#         kill_filters,
#         ["roundNum"],
#         ["attackerName"],
#         [["sum"]],
#         ["RoundNum", "Killers"],
#     )
#     victims = calc_stats(
#         kill_data,
#         kill_filters,
#         ["roundNum"],
#         ["victimName"],
#         [["sum"]],
#         ["RoundNum", "Victims"],
#     )
#     assisters = calc_stats(
#         kill_data.loc[
#             (kill_data["assisterTeam"] != kill_data["victimTeam"])
#             & (kill_data["assisterTeam"].notnull())
#         ],  # .fillna(""),
#         kill_filters,
#         ["roundNum"],
#         ["assisterName"],
#         [["sum"]],
#         ["RoundNum", "Assisters"],
#     )
#     traded = calc_stats(
#         kill_data.loc[
#             (kill_data["attackerTeam"] != kill_data["victimTeam"])
#             & (kill_data["isTrade"] == True)
#             & (kill_data["attackerTeam"].notnull())
#         ],  # .fillna(""),
#         kill_filters,
#         ["roundNum"],
#         ["playerTradedName"],
#         [["sum"]],
#         ["RoundNum", "Traded"],
#     )
#     if flash_assists:
#         flash_assisters = calc_stats(
#             kill_data.loc[
#                 (kill_data["flashThrowerTeam"] != kill_data["victimTeam"])
#                 & (kill_data["flashThrowerTeam"].notnull())
#             ],  # .fillna(""),
#             kill_filters,
#             ["roundNum"],
#             ["flashThrowerName"],
#             [["sum"]],
#             ["RoundNum", "Flash Assisters"],
#         )
#         assisters = assisters.merge(flash_assisters, on="RoundNum")
#         assisters["Assisters"] = assisters["Assisters"] + assisters["Flash Assisters"]
#         assisters = assisters[["RoundNum", "Assisters"]]
#     kast_data = killers.merge(assisters, how="outer")  # .fillna("")
#     kast_data = kast_data.merge(victims, how="outer")  # .fillna("")
#     kast_data = kast_data.merge(traded, how="outer")  # .fillna("")
#     killer_names = kill_data["attackerName"].tolist()
#     cleaned_killer_names = []
#     for x in killer_names:
#         if x != None:
#             cleaned_killer_names.append(x)
#     cleaned_killer_names = pd.Series(cleaned_killer_names)
#     for player in cleaned_killer_names:
#         kast_counts[player] = [[0, 0, 0, 0] for i in range(len(kast_data))]
#         kast_rounds[player] = [0, 0, 0, 0, 0]
#     for rd in kast_data.index:
#         for player in kast_counts:
#             if "K" in kast_string.upper():
#                 kast_counts[player][rd][0] = kast_data.iloc[rd]["Killers"].count(player)
#                 kast_rounds[player][1] += kast_data.iloc[rd]["Killers"].count(player)
#             if "A" in kast_string.upper():
#                 kast_counts[player][rd][1] = kast_data.iloc[rd]["Assisters"].count(
#                     player
#                 )
#                 kast_rounds[player][2] += kast_data.iloc[rd]["Assisters"].count(player)
#             if "S" in kast_string.upper():
#                 if player not in kast_data.iloc[rd]["Victims"]:
#                     kast_counts[player][rd][2] = 1
#                     kast_rounds[player][3] += 1
#             if "T" in kast_string.upper():
#                 kast_counts[player][rd][3] = kast_data.iloc[rd]["Traded"].count(player)
#                 kast_rounds[player][4] += kast_data.iloc[rd]["Traded"].count(player)
#     for player in kast_rounds:
#         for rd in kast_counts[player]:
#             if any(rd):
#                 kast_rounds[player][0] += 1
#         kast_rounds[player][0] /= len(kast_data)
#     kast = pd.DataFrame.from_dict(kast_rounds, orient="index").reset_index()
#     kast.columns = ["Player", f"{kast_string.upper()}%", "K", "A", "S", "T"]
#     kast = kast[columns]
#     kast  # .fillna(0, inplace=True)
#     kast.sort_values(by=f"{kast_string.upper()}%", ascending=False, inplace=True)
#     kast.reset_index(drop=True, inplace=True)
#     return kast


# def kill_stats(
#     damage_data: pd.DataFrame,
#     kill_data: pd.DataFrame,
#     round_data: pd.DataFrame,
#     weapon_fire_data: pd.DataFrame,
#     team: bool = False,
#     damage_filters: Dict[str, Union[List[bool], List[str]]] = {},
#     kill_filters: Dict[str, Union[List[bool], List[str]]] = {},
#     death_filters: Dict[str, Union[List[bool], List[str]]] = {},
#     round_filters: Dict[str, Union[List[bool], List[str]]] = {},
#     weapon_fire_filters: Dict[str, Union[List[bool], List[str]]] = {},
# ) -> pd.DataFrame:
#     """Returns a dataframe with kill statistics.

#     Args:
#         damage_data: A dataframe with damage data.
#         kill_data: A dataframe with kill data.
#         round_data: A dataframe with round data.
#         weapon_fire_data: A dataframe with weapon fire data.
#             where each round is a dictionary.
#         team: A boolean specifying whether to calculate statistics for each team
#             or for each player. The default is to calculate statistics for
#             each player.
#         damage_filters: A dictionary where the keys are the columns of the
#             dataframe represented by damage_data to filter the damage data by
#             and the values are lists that contain the column filters.
#         kill_filters: A dictionary where the keys are the columns of the
#             dataframe represented by kill_data to filter the kill data by and
#             the values are lists that contain the column filters.
#         death_filters: A dictionary where the keys are the columns of the
#             dataframe represented by kill_data to filter the death data by and
#             the values are lists that contain the column filters.
#         round_filters: A dictionary where the keys are the columns of the
#             dataframe represented by round_data to filter the round data by and
#             the values are lists that contain the column filters.
#         weapon_fire_filters: A dictionary where the keys are the columns of the
#             dataframe represented by weapon_fires to filter the weapon fire data
#             by and the values are lists that contain the column filters.

#     Returns:
#         A dataframe with kills, deaths, assists, flash assists, first kills/deaths, headshots, headshot kills/total kills, KAST and accuracy statistics.
#     """
#     stats = ["attackerName", "victimName", "assisterName", "flashThrowerName", "Player"]
#     if team:
#         stats = [
#             "attackerTeam",
#             "victimTeam",
#             "assisterTeam",
#             "flashThrowerTeam",
#             "Team",
#         ]
#     kills = calc_stats(
#         kill_data.loc[kill_data["attackerTeam"] != kill_data["victimTeam"]],
#         kill_filters,
#         [stats[0]],
#         [stats[0]],
#         [["size"]],
#         [stats[4], "K"],
#     )
#     deaths = calc_stats(
#         kill_data,
#         death_filters,
#         [stats[1]],
#         [stats[1]],
#         [["size"]],
#         [stats[4], "D"],
#     )
#     assists = calc_stats(
#         kill_data.loc[kill_data["assisterTeam"] != kill_data["victimTeam"]],
#         kill_filters,
#         [stats[2]],
#         [stats[2]],
#         [["size"]],
#         [stats[4], "A"],
#     )
#     flash_assists = calc_stats(
#         kill_data.loc[kill_data["flashThrowerTeam"] != kill_data["victimTeam"]],
#         kill_filters,
#         [stats[3]],
#         [stats[3]],
#         [["size"]],
#         [stats[4], "FA"],
#     )
#     first_kills = calc_stats(
#         kill_data.loc[
#             (kill_data["attackerTeam"] != kill_data["victimTeam"])
#             & (kill_data["isFirstKill"] == True)
#         ],
#         kill_filters,
#         [stats[0]],
#         [stats[0]],
#         [["size"]],
#         [stats[4], "FK"],
#     )
#     first_deaths = calc_stats(
#         kill_data.loc[
#             (kill_data["attackerTeam"] != kill_data["victimTeam"])
#             & (kill_data["isFirstKill"] == True)
#         ],
#         kill_filters,
#         [stats[1]],
#         [stats[1]],
#         [["size"]],
#         [stats[4], "FD"],
#     )
#     headshots = calc_stats(
#         kill_data.loc[
#             (kill_data["attackerTeam"] != kill_data["victimTeam"])
#             & (kill_data["isHeadshot"] == True)
#         ],
#         kill_filters,
#         [stats[0]],
#         [stats[0]],
#         [["size"]],
#         [stats[4], "HS"],
#     )
#     headshot_pct = calc_stats(
#         kill_data.loc[kill_data["attackerTeam"] != kill_data["victimTeam"]],
#         kill_filters,
#         [stats[0]],
#         ["isHeadshot"],
#         [["mean"]],
#         [stats[4], "HS%"],
#     )
#     if not team:
#         acc_stats = accuracy(
#             damage_data, weapon_fire_data, False, damage_filters, weapon_fire_filters
#         )
#     else:
#         acc_stats = accuracy(
#             damage_data, weapon_fire_data, True, damage_filters, weapon_fire_filters
#         )
#     kast_stats = kast(kill_data, "KAST", kill_filters, death_filters)
#     kill_stats = kills.merge(deaths, how="outer")  # .fillna(0)
#     kill_stats = kill_stats.merge(assists, how="outer")  # .fillna(0)
#     kill_stats = kill_stats.merge(flash_assists, how="outer")  # .fillna(0)
#     kill_stats = kill_stats.merge(first_kills, how="outer")  # .fillna(0)
#     kill_stats = kill_stats.merge(first_deaths, how="outer")  # .fillna(0)
#     kill_stats = kill_stats.merge(headshots, how="outer")  # .fillna(0)
#     kill_stats = kill_stats.merge(headshot_pct, how="outer")  # .fillna(0)
#     kill_stats = kill_stats.merge(acc_stats, how="outer")  # .fillna(0)
#     if not team:
#         kill_stats = kill_stats.merge(kast_stats, how="outer")  # .fillna(0)
#     kill_stats["+/-"] = kill_stats["K"] - kill_stats["D"]
#     kill_stats["KDR"] = kill_stats["K"] / kill_stats["D"]
#     kill_stats["KPR"] = kill_stats["K"] / len(
#         calc_stats(round_data, round_filters, [], [], [], round_data.columns)
#     )
#     kill_stats["FK +/-"] = kill_stats["FK"] - kill_stats["FD"]
#     int_stats = ["K", "D", "A", "FA", "+/-", "FK", "FK +/-", "HS", "T"]
#     if team:
#         int_stats = int_stats[0:-1]
#     kill_stats[int_stats] = kill_stats[int_stats].astype(int)
#     kill_stats["HS%"] = kill_stats["HS%"].astype(float)
#     order = [
#         stats[4],
#         "K",
#         "D",
#         "A",
#         "FA",
#         "+/-",
#         "FK",
#         "FK +/-",
#         "T",
#         "HS",
#         "HS%",
#         "ACC%",
#         "HS ACC%",
#         "KDR",
#         "KPR",
#         "KAST%",
#     ]
#     if team:
#         order = order[0:8] + order[9:-1]
#     kill_stats = kill_stats[order]
#     kill_stats.sort_values(by="K", ascending=False, inplace=True)
#     kill_stats.reset_index(drop=True, inplace=True)
#     return kill_stats


# def adr(
#     damage_data: pd.DataFrame,
#     round_data: pd.DataFrame,
#     team: bool = False,
#     damage_filters: Dict[str, Union[List[bool], List[str]]] = {},
#     round_filters: Dict[str, Union[List[bool], List[str]]] = {},
# ) -> pd.DataFrame:
#     """Returns a dataframe with ADR statistics.

#     Args:
#         damage_data: A dataframe with damage data.
#         round_data: A dataframe with round data.
#         team: A boolean specifying whether to calculate statistics for each team
#             or for each player. The default is to calculate statistics for
#             each player.
#         damage_filters: A dictionary where the keys are the columns of the
#             dataframe represented by damage_data to filter the damage data by
#             and the values are lists that contain the column filters.
#         round_filters: A dictionary where the keys are the columns of the
#             dataframe represented by round_data to filter the round data by and
#             the values are lists that contain the column filters.

#     Returns:
#         A dataframe with the average damage per round (ADR).
#     """
#     stats = ["attackerName", "Player"]
#     if team:
#         stats = ["attackerTeam", "Team"]
#     adr = calc_stats(
#         damage_data.loc[damage_data["attackerTeam"] != damage_data["victimTeam"]],
#         damage_filters,
#         [stats[0]],
#         ["hpDamageTaken", "hpDamage"],
#         [["sum"], ["sum"]],
#         [stats[1], "Norm ADR", "Raw ADR"],
#     )
#     adr["Norm ADR"] = adr["Norm ADR"] / len(
#         calc_stats(round_data, round_filters, [], [], [], round_data.columns)
#     )
#     adr["Raw ADR"] = adr["Raw ADR"] / len(
#         calc_stats(round_data, round_filters, [], [], [], round_data.columns)
#     )
#     adr.sort_values(by="Norm ADR", ascending=False, inplace=True)
#     adr.reset_index(drop=True, inplace=True)
#     return adr


# def rating(
#     damage_data: pd.DataFrame,
#     kill_data: pd.DataFrame,
#     round_data: pd.DataFrame,
#     damage_filters: Dict[str, Union[List[bool], List[str]]] = {},
#     death_filters: Dict[str, Union[List[bool], List[str]]] = {},
#     kill_filters: Dict[str, Union[List[bool], List[str]]] = {},
#     round_filters: Dict[str, Union[List[bool], List[str]]] = {},
# ) -> pd.DataFrame:
#     """Returns a dataframe with an HLTV-esque rating, found by doing:

#     Rating = 0.0073*KAST + 0.3591*KPR + -0.5329*DPR + 0.2372*Impact + 0.0032*ADR + 0.1587
#     where Impact = 2.13*KPR + 0.42*Assist per Round -0.41

#     Args:
#         damage_data: A dataframe with damage data.
#         kill_data: A dataframe with damage data.
#         round_data: A dataframe with round data.
#         damage_filters: A dictionary where the keys are the columns of the
#             dataframe represented by damage_data to filter the damage data by
#             and the values are lists that contain the column filters.
#         death_filters: A dictionary where the keys are the columns of the
#             dataframe represented by kill_data to filter the death data by and
#             the values are lists that contain the column filters.
#         kill_filters: A dictionary where the keys are the columns of the
#             dataframe represented by kill_data to filter the kill data by and
#             the values are lists that contain the column filters.
#         round_filters: A dictionary where the keys are the columns of the
#             dataframe represented by round_data to filter the round data by and
#             the values are lists that contain the column filters.

#     Returns:
#         A dataframe with the impact rating and rating 2.0
#     """
#     kast_stats = kast(kill_data, "KAST", kill_filters, death_filters)
#     kast_stats = kast_stats[["Player", "KAST%"]]
#     kast_stats.columns = ["Player", "KAST"]
#     adr_stats = adr(damage_data, round_data, damage_filters, round_filters)
#     adr_stats = adr_stats[["Player", "Norm ADR"]]
#     adr_stats.columns = ["Player", "ADR"]
#     stats = ["attackerName", "victimName", "assisterName", "flashThrowerName", "Player"]
#     kills = calc_stats(
#         kill_data.loc[kill_data["attackerTeam"] != kill_data["victimTeam"]],
#         kill_filters,
#         [stats[0]],
#         [stats[0]],
#         [["size"]],
#         [stats[4], "K"],
#     )
#     deaths = calc_stats(
#         kill_data,
#         death_filters,
#         [stats[1]],
#         [stats[1]],
#         [["size"]],
#         [stats[4], "D"],
#     )
#     assists = calc_stats(
#         kill_data.loc[kill_data["assisterTeam"] != kill_data["victimTeam"]],
#         kill_filters,
#         [stats[2]],
#         [stats[2]],
#         [["size"]],
#         [stats[4], "A"],
#     )
#     kill_stats = kills.merge(deaths, how="outer")  # .fillna(0)
#     kill_stats = kill_stats.merge(assists, how="outer")  # .fillna(0)
#     kill_stats["KPR"] = kill_stats["K"] / len(
#         calc_stats(round_data, round_filters, [], [], [], round_data.columns)
#     )
#     kill_stats["DPR"] = kill_stats["D"] / len(
#         calc_stats(round_data, round_filters, [], [], [], round_data.columns)
#     )
#     kill_stats["APR"] = kill_stats["A"] / len(
#         calc_stats(round_data, round_filters, [], [], [], round_data.columns)
#     )
#     kill_stats = kill_stats[["Player", "KPR", "DPR", "APR"]]
#     kill_stats = kill_stats.merge(adr_stats, how="outer")  # .fillna(0)
#     kill_stats = kill_stats.merge(kast_stats, how="outer")  # .fillna(0)
#     kill_stats["Impact"] = 2.13 * kill_stats["KPR"] + 0.42 * kill_stats["APR"] - 0.41
#     kill_stats["Rating"] = (
#         0.73 * kill_stats["KAST"]
#         + 0.3591 * kill_stats["KPR"]
#         - 0.5329 * kill_stats["DPR"]
#         + 0.2372 * kill_stats["Impact"]
#         + 0.0032 * kill_stats["ADR"]
#         + 0.1587
#     )
#     kill_stats = kill_stats[["Player", "Impact", "Rating"]]
#     kill_stats.sort_values(by="Rating", ascending=False, inplace=True)
#     kill_stats.reset_index(drop=True, inplace=True)
#     return kill_stats


# def util_dmg(
#     damage_data: pd.DataFrame,
#     grenade_data: pd.DataFrame,
#     team: bool = False,
#     damage_filters: Dict[str, Union[List[bool], List[str]]] = {},
#     grenade_filters: Dict[str, Union[List[bool], List[str]]] = {},
# ) -> pd.DataFrame:
#     """Returns a dataframe with utility damage statistics.

#     Args:
#         damage_data: A dataframe with damage data.
#         grenade_data: A dataframe with grenade data.
#         team: A boolean specifying whether to calculate statistics for each team
#             or for each player. The default is to calculate statistics for
#             each player.
#         damage_filters: A dictionary where the keys are the columns of the
#             dataframe represented by damage_data to filter the damage data by
#             and the values are lists that contain the column filters.
#         grenade_filters: A dictionary where the keys are the columns of the
#             dataframe represented by grenade_data to filter the grenade data by
#             and the values are lists that contain the column filters.

#     Returns:
#         A dataframe with nades thrown, given utility damage (UD) per nade
#     """
#     stats = ["attackerName", "throwerName", "Player"]
#     if team:
#         stats = ["attackerTeam", "throwerTeam", "Team"]
#     util_dmg = calc_stats(
#         damage_data.loc[
#             (damage_data["attackerTeam"] != damage_data["victimTeam"])
#             & (
#                 damage_data["weapon"].isin(
#                     ["HE Grenade", "Incendiary Grenade", "Molotov"]
#                 )
#             )
#         ],
#         damage_filters,
#         [stats[0]],
#         ["hpDamageTaken", "hpDamage"],
#         [["sum"], ["sum"]],
#         [stats[2], "Given UD", "UD"],
#     )
#     nades_thrown = calc_stats(
#         grenade_data.loc[
#             grenade_data["grenadeType"].isin(
#                 ["HE Grenade", "Incendiary Grenade", "Molotov"]
#             )
#         ],
#         grenade_filters,
#         [stats[1]],
#         [stats[1]],
#         [["size"]],
#         [stats[2], "Nades Thrown"],
#     )
#     util_dmg_stats = util_dmg.merge(nades_thrown, how="outer")  # .fillna(0)
#     util_dmg_stats["Given UD Per Nade"] = (
#         util_dmg_stats["Given UD"] / util_dmg_stats["Nades Thrown"]
#     )
#     util_dmg_stats["UD Per Nade"] = (
#         util_dmg_stats["UD"] / util_dmg_stats["Nades Thrown"]
#     )
#     util_dmg_stats.sort_values(by="Given UD", ascending=False, inplace=True)
#     util_dmg_stats.reset_index(drop=True, inplace=True)
#     return util_dmg_stats


# def flash_stats(
#     flash_data: pd.DataFrame,
#     grenade_data: pd.DataFrame,
#     kill_data: pd.DataFrame,
#     team: bool = False,
#     flash_filters: Dict[str, Union[List[bool], List[str]]] = {},
#     grenade_filters: Dict[str, Union[List[bool], List[str]]] = {},
#     kill_filters: Dict[str, Union[List[bool], List[str]]] = {},
# ) -> pd.DataFrame:
#     """Returns a dataframe with flashbang statistics.

#     Args:
#         flash_data: A dataframe with flash data.
#         grenade_data: A dataframe with grenade data.
#         kill_data: A dataframe with kill data.
#         team: A boolean specifying whether to calculate statistics for each team
#             or for each player. The default is to calculate statistics for
#             each player.
#         flash_filters: A dictionary where the keys are the columns of the
#             dataframe represented by flash_data to filter the flash data by
#             and the values are lists that contain the column filters.
#         grenade_filters: A dictionary where the keys are the columns of the
#             dataframe represented by grenade_data to filter the grenade data by
#             and the values are lists that contain the column filters.
#         kill_filters: A dictionary where the keys are the columns of the
#             dataframe represented by kill_data to filter the kill data by and
#             the values are lists that contain the column filters.

#     Returns:
#         A dataframe with the effective flashes (EF), flash assists (FA), effective blind time (EBT), team flashes (TF), total flashes
#     """
#     stats = ["attackerName", "flashThrowerName", "throwerName", "Player"]
#     if team:
#         stats = ["attackerTeam", "flashThrowerTeam", "throwerTeam", "Team"]
#     enemy_flashes = calc_stats(
#         flash_data.loc[flash_data["attackerTeam"] != flash_data["playerTeam"]],
#         flash_filters,
#         [stats[0]],
#         [stats[0]],
#         [["size"]],
#         [stats[3], "EF"],
#     )
#     flash_assists = calc_stats(
#         kill_data.loc[kill_data["flashThrowerTeam"] != kill_data["victimTeam"]],
#         kill_filters,
#         [stats[1]],
#         [stats[1]],
#         [["size"]],
#         [stats[3], "FA"],
#     )
#     blind_time = calc_stats(
#         flash_data.loc[flash_data["attackerTeam"] != flash_data["playerTeam"]],
#         flash_filters,
#         [stats[0]],
#         ["flashDuration"],
#         [["sum"]],
#         [stats[3], "EBT"],
#     )
#     team_flashes = calc_stats(
#         flash_data.loc[flash_data["attackerTeam"] == flash_data["playerTeam"]],
#         flash_filters,
#         [stats[0]],
#         [stats[0]],
#         [["size"]],
#         [stats[3], "TF"],
#     )
#     flashes_thrown = calc_stats(
#         grenade_data.loc[grenade_data["grenadeType"] == "Flashbang"],
#         flash_filters,
#         [stats[2]],
#         [stats[2]],
#         [["size"]],
#         [stats[3], "Flashes Thrown"],
#     )
#     flash_stats = enemy_flashes.merge(flash_assists, how="outer")  # .fillna(0)
#     flash_stats = flash_stats.merge(blind_time, how="outer")  # .fillna(0)
#     flash_stats = flash_stats.merge(team_flashes, how="outer")  # .fillna(0)
#     flash_stats = flash_stats.merge(flashes_thrown, how="outer")  # .fillna(0)
#     flash_stats["EF Per Throw"] = flash_stats["EF"] / flash_stats["Flashes Thrown"]
#     flash_stats["EBT Per Enemy"] = flash_stats["EBT"] / flash_stats["EF"]
#     flash_stats["FA"] = flash_stats["FA"].astype(int)
#     flash_stats.sort_values(by="EF", ascending=False, inplace=True)
#     flash_stats.reset_index(drop=True, inplace=True)
#     return flash_stats


# def bomb_stats(
#     bomb_data: pd.DataFrame,
#     bomb_filters: Dict[str, Union[List[bool], List[str]]] = {},
# ) -> pd.DataFrame:
#     """Returns a dataframe with bomb event statistics.

#     Args:
#         bomb_data: A dataframe with bomb event data.
#         bomb_filters: A dictionary where the keys are the columns of the
#             dataframe represented by bomb_data to filter the bomb data by and
#             the values are lists that contain the column filters.

#     Returns:
#         A dataframe with bomb plant summaries.
#     """
#     team_one = bomb_data["playerTeam"].unique()[0]
#     team_two = bomb_data["playerTeam"].unique()[1]
#     team_one_plants = calc_stats(
#         bomb_data.loc[
#             (bomb_data["bombAction"] == "plant") & (bomb_data["playerTeam"] == team_one)
#         ],
#         bomb_filters,
#         ["bombSite"],
#         ["bombSite"],
#         [["size"]],
#         ["Bombsite", f"{team_one} Plants"],
#     )
#     team_two_plants = calc_stats(
#         bomb_data.loc[
#             (bomb_data["bombAction"] == "plant") & (bomb_data["playerTeam"] == team_two)
#         ],
#         bomb_filters,
#         ["bombSite"],
#         ["bombSite"],
#         [["size"]],
#         ["Bombsite", f"{team_two} Plants"],
#     )
#     team_one_defuses = calc_stats(
#         bomb_data.loc[
#             (bomb_data["bombAction"] == "defuse")
#             & (bomb_data["playerTeam"] == team_one)
#         ],
#         bomb_filters,
#         ["bombSite"],
#         ["bombSite"],
#         [["size"]],
#         ["Bombsite", f"{team_one} Defuses"],
#     )
#     team_two_defuses = calc_stats(
#         bomb_data.loc[
#             (bomb_data["bombAction"] == "defuse")
#             & (bomb_data["playerTeam"] == team_two)
#         ],
#         bomb_filters,
#         ["bombSite"],
#         ["bombSite"],
#         [["size"]],
#         ["Bombsite", f"{team_two} Defuses"],
#     )
#     bomb_stats = team_one_plants.merge(team_two_defuses, how="outer")  # .fillna(0)
#     bomb_stats[f"{team_two} Defuse %"] = (
#         bomb_stats[f"{team_two} Defuses"] / bomb_stats[f"{team_one} Plants"]
#     )
#     bomb_stats = bomb_stats.merge(team_two_plants, how="outer")  # .fillna(0)
#     bomb_stats = bomb_stats.merge(team_one_defuses, how="outer")  # .fillna(0)
#     bomb_stats[f"{team_one} Defuse %"] = (
#         bomb_stats[f"{team_one} Defuses"] / bomb_stats[f"{team_two} Plants"]
#     )
#     bomb_stats.loc[2] = [
#         "A and B",
#         bomb_stats[f"{team_one} Plants"].sum(),
#         bomb_stats[f"{team_two} Defuses"].sum(),
#         (
#             bomb_stats[f"{team_two} Defuses"].sum()
#             / bomb_stats[f"{team_one} Plants"].sum()
#         ),
#         bomb_stats[f"{team_two} Plants"].sum(),
#         bomb_stats[f"{team_one} Defuses"].sum(),
#         (
#             bomb_stats[f"{team_one} Defuses"].sum()
#             / bomb_stats[f"{team_two} Plants"].sum()
#         ),
#     ]
#     bomb_stats  # .fillna(0, inplace=True)
#     bomb_stats.iloc[:, [1, 2, 4, 5]] = bomb_stats.iloc[:, [1, 2, 4, 5]].astype(int)
#     return bomb_stats


# def econ_stats(
#     round_data: pd.DataFrame,
#     round_filters: Dict[str, Union[List[bool], List[str]]] = {},
# ) -> pd.DataFrame:
#     """Returns a dataframe with economy statistics.

#     Args:
#         round_data: A dataframe with round data.
#         round_filters: A dictionary where the keys are the columns of the
#             dataframe represented by round_data to filter the round data by and
#             the values are lists that contain the column filters.

#     Returns:
#         A dataframe with round economy summaries.
#     """
#     ct_stats = calc_stats(
#         round_data,
#         round_filters,
#         ["ctTeam"],
#         ["ctStartEqVal", "ctRoundStartMoney", "ctSpend"],
#         [["mean"], ["mean"], ["mean"]],
#         ["Side", "Avg EQ Value", "Avg Cash", "Avg Spend"],
#     )
#     ct_stats["Side"] = ct_stats["Side"] + " CT"
#     ct_buys = calc_stats(
#         round_data,
#         round_filters,
#         ["ctTeam", "ctBuyType"],
#         ["ctBuyType"],
#         [["size"]],
#         ["Side", "Buy Type", "Counts"],
#     )
#     ct_buys = ct_buys.pivot(index="Side", columns="Buy Type", values="Counts")
#     ct_buys.reset_index(inplace=True)
#     ct_buys.rename_axis(None, axis=1, inplace=True)
#     ct_buys["Side"] = ct_buys["Side"] + " CT"
#     t_stats = calc_stats(
#         round_data,
#         round_filters,
#         ["tTeam"],
#         ["tStartEqVal", "tRoundStartMoney", "tSpend"],
#         [["mean"], ["mean"], ["mean"]],
#         ["Side", "Avg EQ Value", "Avg Cash", "Avg Spend"],
#     )
#     t_stats["Side"] = t_stats["Side"] + " T"
#     t_buys = calc_stats(
#         round_data,
#         round_filters,
#         ["tTeam", "tBuyType"],
#         ["tBuyType"],
#         [["size"]],
#         ["Side", "Buy Type", "Counts"],
#     )
#     t_buys = t_buys.pivot(index="Side", columns="Buy Type", values="Counts")
#     t_buys.reset_index(inplace=True)
#     t_buys.rename_axis(None, axis=1, inplace=True)
#     t_buys["Side"] = t_buys["Side"] + " T"
#     econ_buys = ct_buys.append(t_buys)
#     econ_stats = ct_stats.append(t_stats)
#     econ_stats = econ_buys.merge(econ_stats, how="outer")
#     econ_stats  # .fillna(0, inplace=True)
#     econ_stats.iloc[:, 1:] = econ_stats.iloc[:, 1:].astype(int)
#     return econ_stats


# def weapon_type(weapon: str) -> str:
#     """Returns the weapon type of a weapon.

#     Args:
#         weapon: Weapon name

#     Returns:
#         A string for the weapon type (Knife, Pistol, Shotgun, SMG, Assault Rifle, Machine Gun, Sniper Rifle, Utility)
#     """
#     if weapon in ["Knife"]:
#         return "Melee Kills"
#     elif weapon in [
#         "CZ-75 Auto",
#         "Desert Eagle",
#         "Dual Berettas",
#         "Five-SeveN",
#         "Glock-18",
#         "P2000",
#         "P250",
#         "R8 Revolver",
#         "Tec-9",
#         "USP-S",
#     ]:
#         return "Pistol Kills"
#     elif weapon in ["MAG-7", "Nova", "Sawed-Off", "XM1014"]:
#         return "Shotgun Kills"
#     elif weapon in ["MAC-10", "MP5-SD", "MP7", "MP9", "P90", "PP-Bizon", "UMP-45"]:
#         return "SMG Kills"
#     elif weapon in ["AK-47", "AUG", "FAMAS", "Galil AR", "M4A1-S", "M4A4", "SG 553"]:
#         return "Assault Rifle Kills"
#     elif weapon in ["M249", "Negev"]:
#         return "Machine Gun Kills"
#     elif weapon in ["AWP", "G3SG1", "SCAR-20", "SSG 08"]:
#         return "Sniper Rifle Kills"
#     else:
#         return "Utility Kills"


# def kill_breakdown(
#     kill_data: pd.DataFrame,
#     team: bool = False,
#     kill_filters: Dict[str, Union[List[bool], List[str]]] = {},
# ) -> pd.DataFrame:
#     """Returns a dataframe with kills broken down by weapon type.

#     Args:
#         kill_data: A dataframe with kill data.
#         team: A boolean specifying whether to calculate statistics for each team
#             or for each player. The default is to calculate statistics for
#             each player.
#         kill_filters: A dictionary where the keys are the columns of the
#             dataframe represented by kill_data to filter the kill data by and
#             the values are lists that contain the column filters.

#     Returns:
#         A dataframe with kill type statistics.
#     """
#     stats = ["attackerName", "Player"]
#     if team:
#         stats = ["attackerTeam", "Team"]
#     kill_breakdown = kill_data.loc[
#         kill_data["attackerTeam"] != kill_data["victimTeam"]
#     ].copy()
#     kill_breakdown["Kills Type"] = kill_breakdown.apply(
#         lambda row: weapon_type(row["weapon"]), axis=1
#     )
#     kill_breakdown = calc_stats(
#         kill_breakdown,
#         kill_filters,
#         [stats[0], "Kills Type"],
#         [stats[0]],
#         [["size"]],
#         [stats[1], "Kills Type", "Kills"],
#     )
#     kill_breakdown = kill_breakdown.pivot(
#         index=stats[1], columns="Kills Type", values="Kills"
#     )
#     for col in [
#         "Melee Kills",
#         "Pistol Kills",
#         "Shotgun Kills",
#         "SMG Kills",
#         "Assault Rifle Kills",
#         "Machine Gun Kills",
#         "Sniper Rifle Kills",
#         "Utility Kills",
#     ]:
#         if not col in kill_breakdown.columns:
#             kill_breakdown.insert(0, col, 0)
#         kill_breakdown[col]  # .fillna(0, inplace=True)
#         kill_breakdown[col] = kill_breakdown[col].astype(int)
#     kill_breakdown["Total Kills"] = kill_breakdown.iloc[0:].sum(axis=1)
#     kill_breakdown.reset_index(inplace=True)
#     kill_breakdown.rename_axis(None, axis=1, inplace=True)
#     kill_breakdown = kill_breakdown[
#         [
#             stats[1],
#             "Melee Kills",
#             "Pistol Kills",
#             "Shotgun Kills",
#             "SMG Kills",
#             "Assault Rifle Kills",
#             "Machine Gun Kills",
#             "Sniper Rifle Kills",
#             "Utility Kills",
#             "Total Kills",
#         ]
#     ]
#     kill_breakdown.sort_values(by="Total Kills", ascending=False, inplace=True)
#     kill_breakdown.reset_index(drop=True, inplace=True)
#     return kill_breakdown


# def util_dmg_breakdown(
#     damage_data: pd.DataFrame,
#     grenade_data: pd.DataFrame,
#     team: bool = False,
#     damage_filters: Dict[str, Union[List[bool], List[str]]] = {},
#     grenade_filters: Dict[str, Union[List[bool], List[str]]] = {},
# ) -> pd.DataFrame:
#     """Returns a dataframe with utility damage statistics broken down by grenade type.

#     Args:
#         damage_data: A dataframe with damage data.
#         grenade_data: A dataframe with grenade data.
#         team: A boolean specifying whether to calculate statistics for each team
#             or for each player. The default is to calculate statistics for
#             each player.
#         damage_filters: A dictionary where the keys are the columns of the
#             dataframe represented by damage_data to filter the damage data by
#             and the values are lists that contain the column filters.
#         grenade_filters: A dictionary where the keys are the columns of the
#             dataframe represented by grenade_data to filter the grenade data by
#             and the values are lists that contain the column filters.

#     Returns:
#         A dataframe with utility damage summary.
#     """
#     stats = ["attackerName", "throwerName", "Player"]
#     if team:
#         stats = ["attackerTeam", "throwerTeam", "Team"]
#     util_dmg = calc_stats(
#         damage_data.loc[
#             (damage_data["attackerTeam"] != damage_data["victimTeam"])
#             & (
#                 damage_data["weapon"].isin(
#                     ["HE Grenade", "Incendiary Grenade", "Molotov"]
#                 )
#             )
#         ],
#         damage_filters,
#         [stats[0], "weapon"],
#         ["hpDamageTaken", "hpDamage"],
#         [["sum"], ["sum"]],
#         [stats[2], "Nade Type", "Given UD", "UD"],
#     )
#     nades_thrown = calc_stats(
#         grenade_data.loc[
#             grenade_data["grenadeType"].isin(
#                 ["HE Grenade", "Incendiary Grenade", "Molotov"]
#             )
#         ],
#         grenade_filters,
#         [stats[1], "grenadeType"],
#         [stats[1]],
#         [["size"]],
#         [stats[2], "Nade Type", "Nades Thrown"],
#     )
#     util_dmg_breakdown = util_dmg.merge(
#         nades_thrown, how="outer", on=[stats[2], "Nade Type"]
#     )  # .fillna(0)
#     util_dmg_breakdown["Given UD Per Nade"] = (
#         util_dmg_breakdown["Given UD"] / util_dmg_breakdown["Nades Thrown"]
#     )
#     util_dmg_breakdown["UD Per Nade"] = (
#         util_dmg_breakdown["UD"] / util_dmg_breakdown["Nades Thrown"]
#     )
#     util_dmg_breakdown.sort_values(
#         by=[stats[2], "Given UD"], ascending=[True, False], inplace=True
#     )
#     util_dmg_breakdown.reset_index(drop=True, inplace=True)
#     return util_dmg_breakdown


# def win_breakdown(
#     round_data: pd.DataFrame,
#     round_filters: Dict[str, Union[List[bool], List[str]]] = {},
# ) -> pd.DataFrame:
#     """Returns a dataframe with wins broken down by round end reason.

#     Args:
#         round_data: A dataframe with round data.
#         round_filters: A dictionary where the keys are the columns of the
#             dataframe represented by round_data to filter the round data by and
#             the values are lists that contain the column filters.

#     Returns:
#         A dataframe with round wins described by reason.
#     """
#     round_data_copy = round_data.copy()
#     round_data_copy.replace("BombDefused", "CT Bomb Defusal Wins", inplace=True)
#     round_data_copy.replace("CTWin", "CT T Elim Wins", inplace=True)
#     round_data_copy.replace("TargetBombed", "T Bomb Detonation Wins", inplace=True)
#     round_data_copy.replace("TargetSaved", "CT Time Expired Wins", inplace=True)
#     round_data_copy.replace("TerroristsWin", "T CT Elim Wins", inplace=True)
#     win_breakdown = calc_stats(
#         round_data_copy,
#         round_filters,
#         ["winningTeam", "roundEndReason"],
#         ["roundEndReason"],
#         [["size"]],
#         ["Team", "RoundEndReason", "Count"],
#     )
#     win_breakdown = win_breakdown.pivot(
#         index="Team", columns="RoundEndReason", values="Count"
#     )  # .fillna(0)
#     win_breakdown.reset_index(inplace=True)
#     win_breakdown.rename_axis(None, axis=1, inplace=True)
#     win_breakdown["Total CT Wins"] = win_breakdown.iloc[0:][
#         list(
#             set.intersection(
#                 set(win_breakdown.columns),
#                 set(["CT Bomb Defusal Wins", "CT T Elim Wins", "CT Time Expired Wins"]),
#             )
#         )
#     ].sum(axis=1)
#     win_breakdown["Total T Wins"] = win_breakdown.iloc[0:][
#         list(
#             set.intersection(
#                 set(win_breakdown.columns),
#                 set(["T Bomb Detonation Wins", "T CT Elim Wins"]),
#             )
#         )
#     ].sum(axis=1)
#     win_breakdown["Total Wins"] = win_breakdown.iloc[0:, 0:-2].sum(axis=1)
#     win_breakdown.iloc[:, 1:] = win_breakdown.iloc[:, 1:].astype(int)
#     return win_breakdown


# def player_box_score(
#     damage_data: pd.DataFrame,
#     flash_data: pd.DataFrame,
#     grenade_data: pd.DataFrame,
#     kill_data: pd.DataFrame,
#     round_data: pd.DataFrame,
#     weapon_fire_data: pd.DataFrame,
#     damage_filters: Dict[str, Union[List[bool], List[str]]] = {},
#     flash_filters: Dict[str, Union[List[bool], List[str]]] = {},
#     grenade_filters: Dict[str, Union[List[bool], List[str]]] = {},
#     kill_filters: Dict[str, Union[List[bool], List[str]]] = {},
#     death_filters: Dict[str, Union[List[bool], List[str]]] = {},
#     round_filters: Dict[str, Union[List[bool], List[str]]] = {},
#     weapon_fire_filters: Dict[str, Union[List[bool], List[str]]] = {},
# ) -> pd.DataFrame:
#     """Returns a player box score dataframe.

#     Args:
#        damage_data: A dataframe with damage data.
#        flash_data: A dataframe with flash data.
#        grenade_data: A dataframe with grenade data.
#        kill_data: A dataframe with kill data.
#        round_data: A dataframe with round data.
#        weapon_fire_data: A dataframe with weapon fire data.
#        damage_filters: A dictionary where the keys are the columns of the
#            dataframe represented by damage_data to filter the damage data by
#            and the values are lists that contain the column filters.
#        flash_filters: A dictionary where the keys are the columns of the
#            dataframe represented by flash_data to filter the flash data by
#            and the values are lists that contain the column filters.
#        grenade_filters: A dictionary where the keys are the columns of the
#            dataframe represented by grenade_data to filter the grenade data by
#            and the values are lists that contain the column filters.
#        kill_filters: A dictionary where the keys are the columns of the
#            dataframe represented by kill_data to filter the kill data by and
#            the values are lists that contain the column filters.
#        death_filters: A dictionary where the keys are the columns of the
#            dataframe represented by kill_data to filter the death data by and
#            the values are lists that contain the column filters.
#        round_filters: A dictionary where the keys are the columns of the
#            dataframe represented by round_data to filter the round data by and
#            the values are lists that contain the column filters.
#        weapon_fire_filters: A dictionary where the keys are the columns of the
#            dataframe to filter the weapon fire data by and the values are lists
#            that contain the column filters.

#     Returns:
#         A dataframe with player summaries.
#     """
#     k_stats = kill_stats(
#         damage_data,
#         kill_data,
#         round_data,
#         weapon_fire_data,
#         damage_filters,
#         kill_filters,
#         death_filters,
#         round_filters,
#         weapon_fire_filters,
#     )
#     k_stats = k_stats[
#         ["Player", "K", "D", "A", "FA", "HS%", "ACC%", "HS ACC%", "KDR", "KAST%"]
#     ]
#     adr_stats = adr(damage_data, round_data, damage_filters, round_filters)
#     adr_stats = adr_stats[["Player", "Norm ADR"]]
#     adr_stats.columns = ["Player", "ADR"]
#     ud_stats = util_dmg(damage_data, grenade_data, damage_filters, grenade_filters)
#     ud_stats = ud_stats[["Player", "UD", "UD Per Nade"]]
#     f_stats = flash_stats(
#         flash_data,
#         grenade_data,
#         kill_data,
#         flash_filters,
#         grenade_filters,
#         kill_filters,
#     )
#     f_stats = f_stats[["Player", "EF", "EF Per Throw"]]
#     rating_stats = rating(
#         damage_data,
#         kill_data,
#         round_data,
#         damage_filters,
#         death_filters,
#         kill_filters,
#         round_filters,
#     )
#     box_score = k_stats.merge(adr_stats, how="outer")  # .fillna(0)
#     box_score = box_score.merge(ud_stats, how="outer")  # .fillna(0)
#     box_score = box_score.merge(f_stats, how="outer")  # .fillna(0)
#     box_score = box_score.merge(rating_stats, how="outer")  # .fillna(0)
#     return box_score


# def team_box_score(
#     damage_data: pd.DataFrame,
#     flash_data: pd.DataFrame,
#     grenade_data: pd.DataFrame,
#     kill_data: pd.DataFrame,
#     round_data: pd.DataFrame,
#     weapon_fire_data: pd.DataFrame,
#     damage_filters: Dict[str, Union[List[bool], List[str]]] = {},
#     flash_filters: Dict[str, Union[List[bool], List[str]]] = {},
#     grenade_filters: Dict[str, Union[List[bool], List[str]]] = {},
#     kill_filters: Dict[str, Union[List[bool], List[str]]] = {},
#     death_filters: Dict[str, Union[List[bool], List[str]]] = {},
#     round_filters: Dict[str, Union[List[bool], List[str]]] = {},
#     weapon_fire_filters: Dict[str, Union[List[bool], List[str]]] = {},
# ) -> pd.DataFrame:
#     """Returns a team box score dataframe.

#     Args:
#        damage_data: A dataframe with damage data.
#        flash_data: A dataframe with flash data.
#        grenade_data: A dataframe with grenade data.
#        kill_data: A dataframe with kill data.
#        round_data: A dataframe with round data.
#        weapon_fire_data: A dataframe with weapon fire data.
#        damage_filters: A dictionary where the keys are the columns of the
#            dataframe represented by damage_data to filter the damage data by
#            and the values are lists that contain the column filters.
#        flash_filters: A dictionary where the keys are the columns of the
#            dataframe represented by flash_data to filter the flash data by
#            and the values are lists that contain the column filters.
#        grenade_filters: A dictionary where the keys are the columns of the
#            dataframe represented by grenade_data to filter the grenade data by
#            and the values are lists that contain the column filters.
#        kill_filters: A dictionary where the keys are the columns of the
#            dataframe represented by kill_data to filter the kill data by and
#            the values are lists that contain the column filters.
#        death_filters: A dictionary where the keys are the columns of the
#            dataframe represented by kill_data to filter the death data by and
#            the values are lists that contain the column filters.
#        round_filters: A dictionary where the keys are the columns of the
#            dataframe represented by round_data to filter the round data by and
#            the values are lists that contain the column filters.
#        weapon_fire_filters: A dictionary where the keys are the columns of the
#            dataframe to filter the weapon fire data by and the values are lists
#            that contain the column filters.

#     Returns:
#         A dataframe with team summaries.
#     """
#     k_stats = kill_stats(
#         damage_data,
#         kill_data,
#         round_data,
#         weapon_fire_data,
#         True,
#         damage_filters,
#         kill_filters,
#         death_filters,
#         round_filters,
#         weapon_fire_filters,
#     )
#     acc_stats = accuracy(
#         damage_data, weapon_fire_data, True, damage_filters, weapon_fire_filters
#     )
#     adr_stats = adr(damage_data, round_data, True, damage_filters, round_filters)
#     ud_stats = util_dmg(
#         damage_data, grenade_data, True, damage_filters, grenade_filters
#     )
#     f_stats = flash_stats(
#         flash_data,
#         grenade_data,
#         kill_data,
#         True,
#         flash_filters,
#         grenade_filters,
#         kill_filters,
#     )
#     e_stats = econ_stats(round_data, round_filters)
#     for index in e_stats.index:
#         e_stats.iloc[index, 0] = e_stats["Side"].str.rsplit(n=1)[index][0]
#         rounds = e_stats.iloc[index, 1:-4].sum()
#         e_stats.iloc[index, -3:] = e_stats.iloc[index, -3:] * rounds
#     e_stats = e_stats.groupby(["Side"]).sum()
#     e_stats.reset_index(inplace=True)
#     e_stats.iloc[:, -3:] = (
#         e_stats.iloc[:, -3:] / len(filter_df(round_data, round_filters))
#     ).astype(int)
#     e_stats.rename(columns={"Side": "Team"}, inplace=True)
#     box_score = k_stats.merge(acc_stats, how="outer")
#     box_score = box_score.merge(adr_stats, how="outer")
#     box_score = box_score.merge(ud_stats, how="outer")
#     box_score = box_score.merge(f_stats, how="outer")
#     box_score = box_score.merge(e_stats, how="outer")
#     box_score = box_score.merge(
#         win_breakdown(round_data, round_filters), how="outer"
#     )  # .fillna(0)
#     box_score.rename(
#         columns={
#             "Norm ADR": "ADR",
#             "Total CT Wins": "CT Wins",
#             "Total T Wins": "T Wins",
#             "Total Wins": "Score",
#         },
#         inplace=True,
#     )
#     score = box_score["Score"]
#     ct_wins = box_score["CT Wins"]
#     t_wins = box_score["T Wins"]
#     box_score.drop(["Score", "CT Wins", "T Wins"], axis=1, inplace=True)
#     box_score.insert(1, "Score", score)
#     box_score.insert(2, "CT Wins", ct_wins)
#     box_score.insert(3, "T Wins", t_wins)
#     box_score = box_score.transpose()
#     box_score.columns = box_score.iloc[0]
#     box_score.drop("Team", inplace=True)
#     box_score.rename_axis(None, axis=1, inplace=True)
#     box_score = box_score.loc[
#         [
#             "Score",
#             "CT Wins",
#             "T Wins",
#             "K",
#             "D",
#             "A",
#             "FA",
#             "+/-",
#             "FK",
#             "HS",
#             "HS%",
#             "Strafe%",
#             "ACC%",
#             "HS ACC%",
#             "ADR",
#             "UD",
#             "Nades Thrown",
#             "UD Per Nade",
#             "EF",
#             "Flashes Thrown",
#             "EF Per Throw",
#             "EBT Per Enemy",
#         ],
#         :,
#     ].append(box_score.iloc[31:, :])
#     return box_score



import json
import os
import pandas as pd
import numpy as np
import networkx as nx

from awpy.utils import transform_csv_to_json
from pathlib import Path

path = os.path.join(os.path.dirname(__file__), "")

# Create nav tile info
nav_dfs = []
for file in os.listdir(path + "nav/"):
    if file.endswith(".csv"):
        df = pd.read_csv(path + "nav/" + file)
        nav_dfs.append(df)

NAV_CSV = pd.concat(nav_dfs, ignore_index=True)
NAV_CSV.areaName = NAV_CSV.areaName.fillna("")
NAV = transform_csv_to_json(NAV_CSV)

# Create nav graphs
NAV_GRAPHS = {}
for m in NAV.keys():
    G = nx.Graph()
    for a in NAV[m].keys():
        r = NAV[m][a]
        G.add_nodes_from(
            [
                (
                    a,
                    {
                        "mapName": m,
                        "areaID": a,
                        "areaName": r["areaName"],
                        "northWestX": r["northWestX"],
                        "northWestY": r["northWestY"],
                        "northWestZ": r["northWestZ"],
                        "southEastX": r["southEastX"],
                        "southEastY": r["southEastY"],
                        "southEastZ": r["southEastZ"],
                        "size": np.sqrt(
                            (r["northWestX"] - r["southEastX"]) ** 2
                            + (r["northWestY"] - r["southEastY"]) ** 2
                            + (r["northWestZ"] - r["southEastZ"]) ** 2
                        ),
                    },
                ),
            ]
        )
    edge_list = open(path + "nav/" + m + ".txt", "r")
    edge_list_lines = edge_list.readlines()
    for line in edge_list_lines:
        areas = line.strip().split(",")
        G.add_edge(int(areas[0]), int(areas[1]))
    NAV_GRAPHS[m] = G

# Open map data
with open(Path(path + "map/map_data.json")) as f:
    MAP_DATA = json.load(f)



import json
import logging
import os
import subprocess
import pandas as pd

from awpy.utils import check_go_version


class DemoParser:
    """DemoParser can parse, load and clean data from a CSGO demofile. Can be instantiated without a specified demofile.

    Attributes:
        demofile (string): A string denoting the path to the demo file, which ends in .dem
        demo_id (string): A unique demo name/game id. Default is inferred from demofile name
        log (boolean): A boolean indicating if the log should print to stdout.
        parse_rate (int): One of 128, 64, 32, 16, 8, 4, 2, or 1. The lower the value, the more frames are collected. Indicates spacing between parsed demo frames in ticks. Default is 128.
        parse_frames (bool): Flag if you want to parse frames (trajectory data) or not
        parse_kill_frames (bool): Flag if you want to parse frames on kills
        trade_time (int): Length of the window for a trade (in seconds). Default is 5.
        dmg_rolled (bool): Boolean if you want damages rolled up (since multiple damages for a player can happen in 1 tick from the same weapon.)
        buy_style (string): Buy style string, one of "hltv" or "csgo"

    Raises:
        ValueError: Raises a ValueError if the Golang version is lower than 1.17
    """

    def __init__(
        self,
        demofile="",
        outpath=None,
        demo_id=None,
        log=False,
        parse_rate=128,
        parse_frames=True,
        parse_kill_frames=False,
        trade_time=5,
        dmg_rolled=False,
        buy_style="hltv",
        json_indentation=False,
    ):
        # Set up logger
        logging.basicConfig(
            level=logging.INFO,
            format="%(asctime)s [%(levelname)s] %(message)s",
            datefmt="%H:%M:%S",
        )
        self.logger = logging.getLogger("awpy")
        self.logger.propagate = log

        # Handle demofile and demo_id name. Finds right most '/' in case demofile is a specified path.
        self.demofile = os.path.abspath(demofile)
        self.logger.info("Initialized awpy DemoParser with demofile " + self.demofile)
        if (demo_id is None) | (demo_id == ""):
            self.demo_id = demofile[demofile.rfind("/") + 1 : -4]
        else:
            self.demo_id = demo_id
        if outpath is None:
            self.outpath = os.path.abspath(os.getcwd())
        else:
            self.outpath = os.path.abspath(outpath)
        self.logger.info("Setting demo id to " + self.demo_id)

        # Handle parse rate. If the parse rate is less than 64, likely to be slow
        if parse_rate < 1 or type(parse_rate) is not int:
            self.logger.warning(
                "Parse rate of "
                + str(parse_rate)
                + " not acceptable! Parse rate must be an integer greater than 0."
            )
            parse_rate = 128
            self.parse_rate = parse_rate

        if parse_rate < 64 and parse_rate > 1:
            self.logger.warning(
                "A parse rate lower than 64 may be slow depending on the tickrate of the demo, which is usually 64 for MM and 128 for pro demos."
            )
            self.parse_rate = parse_rate
        elif parse_rate >= 256:
            self.logger.warning(
                "A high parse rate means very few frames. Only use for testing purposes."
            )
            self.parse_rate = parse_rate
        else:
            self.parse_rate = parse_rate
        self.logger.info("Setting parse rate to " + str(self.parse_rate))

        # Handle trade time
        if trade_time <= 0:
            self.logger.warning(
                "Trade time can't be negative, setting to default value of 5 seconds."
            )
            self.trade_time = 5
        elif trade_time > 7:
            self.logger.warning(
                "Trade time of "
                + str(trade_time)
                + " is rather long. Consider a value between 4-7."
            )
        else:
            self.trade_time = trade_time
        self.logger.info("Setting trade time to " + str(self.trade_time))

        # Handle buy style
        if buy_style not in ["hltv", "csgo"]:
            self.logger.warning(
                "Buy style specified is not one of hltv, csgo, will be set to hltv by default"
            )
            self.buy_style = "hltv"
        else:
            self.buy_style = buy_style
        self.logger.info("Setting buy style to " + str(self.buy_style))

        self.dmg_rolled = dmg_rolled
        self.parse_frames = parse_frames
        self.parse_kill_frames = parse_kill_frames
        self.json_indentation = json_indentation
        self.logger.info("Rollup damages set to " + str(self.dmg_rolled))
        self.logger.info("Parse frames set to " + str(self.parse_frames))
        self.logger.info("Parse kill frames set to " + str(self.parse_kill_frames))
        self.logger.info("Output json indentation set to " + str(self.json_indentation))

        # Set parse error to False
        self.parse_error = False

    def parse_demo(self):
        """Parse a demofile using the Go script parse_demo.go -- this function needs the .demofile to be set in the class, and the file needs to exist.

        Returns:
            Outputs a JSON file to current working directory.

        Raises:
            ValueError: Raises a ValueError if the Golang version is lower than 1.17
            FileNotFoundError: Raises a FileNotFoundError if the demofile path does not exist.
        """
        # Check if Golang version is compatible
        acceptable_go = check_go_version()
        if not acceptable_go:
            self.logger.error(
                "Error calling Go. Check if Go is installed using 'go version'. Need at least v1.17.0."
            )
            raise ValueError(
                "Error calling Go. Check if Go is installed using 'go version'. Need at least v1.17.0."
            )
        else:
            self.logger.info("Go version>=1.17.0")

        # Check if demofile exists
        if not os.path.exists(os.path.abspath(self.demofile)):
            self.logger.error("Demofile path does not exist!")
            raise FileNotFoundError("Demofile path does not exist!")

        path = os.path.join(os.path.dirname(__file__), "")
        self.logger.info("Running Golang parser from " + path)
        self.logger.info("Looking for file at " + self.demofile)
        self.parser_cmd = [
            "go",
            "run",
            "parse_demo.go",
            "-demo",
            self.demofile,
            "-parserate",
            str(self.parse_rate),
            "-tradetime",
            str(self.trade_time),
            "-buystyle",
            str(self.buy_style),
            "-demoid",
            str(self.demo_id),
            "-out",
            self.outpath,
        ]
        if self.dmg_rolled:
            self.parser_cmd.append("--dmgrolled")
        if self.parse_frames:
            self.parser_cmd.append("--parseframes")
        if self.parse_kill_frames:
            self.parser_cmd.append("--parsekillframes")
        if self.json_indentation:
            self.parser_cmd.append("--jsonindentation")
        proc = subprocess.Popen(
            self.parser_cmd,
            stdout=subprocess.PIPE,
            cwd=path,
        )
        stdout = proc.stdout.read().splitlines()
        self.output_file = self.demo_id + ".json"
        if os.path.isfile(self.outpath + "/" + self.output_file):
            self.logger.info("Wrote demo parse output to " + self.output_file)
            self.parse_error = False
        else:
            self.parse_error = True
            self.logger.error("No file produced, error in calling Golang")
            self.logger.error(stdout)

    def read_json(self, json_path):
        """Reads the JSON file given a JSON path. Can be used to read in already processed demofiles.

        Args:
            json_path (string): Path to JSON file

        Returns:
            JSON in Python dictionary form

        Raises:
            FileNotFoundError: Raises a FileNotFoundError if the JSON path doesn't exist
        """
        # Check if JSON exists
        if not os.path.exists(json_path):
            self.logger.error("JSON path does not exist!")
            raise FileNotFoundError("JSON path does not exist!")

        # Read in json to .json attribute
        with open(json_path, encoding="utf8") as f:
            demo_data = json.load(f)

        self.json = demo_data
        self.logger.info(
            "JSON data loaded, available in the `json` attribute to parser"
        )
        return demo_data

    def parse(self, return_type="json", clean=True):
        """Wrapper for parse_demo() and read_json(). Use to parse a demo.

        Args:
            return_type (string): Either "json" or "df"
            clean (bool, optional): True to run clean_rounds, otherwise, uncleaned data is returned. Defaults to True.

        Returns:
            A dictionary of output (which is parsed to a JSON file in the working directory)

        Raises:
            ValueError: Raises a ValueError if the return_type is not "json" or "df"
            AttributeError: Raises an AttributeError if the .json attribute is None
        """
        self.parse_demo()
        self.read_json(json_path=self.outpath + "/" + self.output_file)
        if clean:
            clean_data = self.clean_rounds()
        if self.json:
            self.logger.info("JSON output found")
            if return_type == "json":
                return self.json
            elif return_type == "df":
                demo_data = self.parse_json_to_df()
                self.logger.info("Returned dataframe output")
                return demo_data
            else:
                self.logger.error("Parse return_type must be either 'json' or 'df'")
                raise ValueError("return_type must be either 'json' or 'df'")
        else:
            self.logger.error("JSON couldn't be returned")
            raise AttributeError("No JSON parsed! Error in producing JSON.")

    def parse_json_to_df(self):
        """Returns JSON into dictionary where keys correspond to data frames

        Returns:
            A dictionary of output

        Raises:
            AttributeError: Raises an AttributeError if the .json attribute is None
        """
        if self.json:
            demo_data = {}
            demo_data["matchID"] = self.json["matchID"]
            demo_data["clientName"] = self.json["clientName"]
            demo_data["mapName"] = self.json["mapName"]
            demo_data["tickRate"] = self.json["tickRate"]
            demo_data["playbackTicks"] = self.json["playbackTicks"]
            # Rounds
            demo_data["rounds"] = self._parse_rounds()
            # Kills
            demo_data["kills"] = self._parse_kills()
            demo_data["kills"]["attackerSteamID"] = demo_data["kills"][
                "attackerSteamID"
            ].astype(pd.Int64Dtype())
            demo_data["kills"]["victimSteamID"] = demo_data["kills"][
                "victimSteamID"
            ].astype(pd.Int64Dtype())
            demo_data["kills"]["assisterSteamID"] = demo_data["kills"][
                "assisterSteamID"
            ].astype(pd.Int64Dtype())
            demo_data["kills"]["flashThrowerSteamID"] = demo_data["kills"][
                "flashThrowerSteamID"
            ].astype(pd.Int64Dtype())
            # Damages
            demo_data["damages"] = self._parse_damages()
            demo_data["damages"]["attackerSteamID"] = demo_data["damages"][
                "attackerSteamID"
            ].astype(pd.Int64Dtype())
            demo_data["damages"]["victimSteamID"] = demo_data["damages"][
                "victimSteamID"
            ].astype(pd.Int64Dtype())
            # Grenades
            demo_data["grenades"] = self._parse_grenades()
            demo_data["grenades"]["throwerSteamID"] = demo_data["grenades"][
                "throwerSteamID"
            ].astype(pd.Int64Dtype())
            # Flashes
            demo_data["flashes"] = self._parse_flashes()
            demo_data["flashes"]["attackerSteamID"] = demo_data["flashes"][
                "attackerSteamID"
            ].astype(pd.Int64Dtype())
            demo_data["flashes"]["playerSteamID"] = demo_data["flashes"][
                "playerSteamID"
            ].astype(pd.Int64Dtype())
            # Weapon Fires
            demo_data["weaponFires"] = self._parse_weapon_fires()
            demo_data["weaponFires"]["playerSteamID"] = demo_data["flashes"][
                "playerSteamID"
            ].astype(pd.Int64Dtype())
            # Bomb Events
            demo_data["bombEvents"] = self._parse_bomb_events()
            demo_data["bombEvents"]["playerSteamID"] = demo_data["bombEvents"][
                "playerSteamID"
            ].astype(pd.Int64Dtype())
            # Frames
            demo_data["frames"] = self._parse_frames()
            # Player Frames
            demo_data["playerFrames"] = self._parse_player_frames()
            demo_data["playerFrames"]["steamID"] = demo_data["playerFrames"][
                "steamID"
            ].astype(pd.Int64Dtype())
            self.logger.info("Returned dataframe output")
            return demo_data
        else:
            self.logger.error(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )
            raise AttributeError(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )

    def _parse_frames(self):
        """Returns frames as a Pandas dataframe

        Returns:
            A Pandas dataframe where each row is a frame (game state) in the demo, which is a discrete point of time.

        Raises:
            AttributeError: Raises an AttributeError if the .json attribute is None
        """
        if self.json:
            frames_dataframes = []
            keys = ["tick", "seconds"]
            for r in self.json["gameRounds"]:
                if r["frames"]:
                    for frame in r["frames"]:
                        frame_item = {}
                        frame_item["roundNum"] = r["roundNum"]
                        for k in keys:
                            frame_item[k] = frame[k]
                        for side in ["ct", "t"]:
                            if side == "ct":
                                frame_item["ctTeamName"] = frame["ct"]["teamName"]
                                frame_item["ctEqVal"] = frame["ct"]["teamEqVal"]
                                frame_item["ctAlivePlayers"] = frame["ct"][
                                    "alivePlayers"
                                ]
                                frame_item["ctUtility"] = frame["ct"]["totalUtility"]
                            else:
                                frame_item["tTeamName"] = frame["t"]["teamName"]
                                frame_item["tEqVal"] = frame["t"]["teamEqVal"]
                                frame_item["tAlivePlayers"] = frame["t"]["alivePlayers"]
                                frame_item["tUtility"] = frame["t"]["totalUtility"]
                        frames_dataframes.append(frame_item)
            frames_df = pd.DataFrame(frames_dataframes)
            frames_df["matchID"] = self.json["matchID"]
            frames_df["mapName"] = self.json["mapName"]
            return pd.DataFrame(frames_dataframes)
        else:
            self.logger.error(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )
            raise AttributeError(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )

    def _parse_player_frames(self):
        """Returns player frames as a Pandas dataframe.

        Returns:
            A Pandas dataframe where each row is a player's attributes at a given frame (game state).

        Raises:
            AttributeError: Raises an AttributeError if the .json attribute is None
        """
        if self.json:
            player_frames = []
            for r in self.json["gameRounds"]:
                if r["frames"]:
                    for frame in r["frames"]:
                        for side in ["ct", "t"]:
                            if frame[side]["players"] is not None and (
                                len(frame[side]["players"])
                                > 0  # Used to be == 5, to ensure the sides were equal.
                            ):
                                for player in frame[side]["players"]:
                                    player_item = {}
                                    player_item["roundNum"] = r["roundNum"]
                                    player_item["tick"] = frame["tick"]
                                    player_item["seconds"] = frame["seconds"]
                                    player_item["side"] = side
                                    player_item["teamName"] = frame[side]["teamName"]
                                    for col in player.keys():
                                        if col != "inventory":
                                            player_item[col] = player[col]
                                    player_frames.append(player_item)
            player_frames_df = pd.DataFrame(player_frames)
            player_frames_df["matchID"] = self.json["matchID"]
            player_frames_df["mapName"] = self.json["mapName"]
            return pd.DataFrame(player_frames_df)
        else:
            self.logger.error(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )
            raise AttributeError(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )

    def _parse_rounds(self):
        """Returns rounds as a Pandas dataframe

        Returns:
            A Pandas dataframe where each row is a round

        Raises:
            AttributeError: Raises an AttributeError if the .json attribute is None
        """
        if self.json:
            rounds = []
            cols = [
                "roundNum",
                "startTick",
                "freezeTimeEndTick",
                "endTick",
                "endOfficialTick",
                "tScore",
                "ctScore",
                "endTScore",
                "endCTScore",
                "tTeam",
                "ctTeam",
                "winningSide",
                "winningTeam",
                "losingTeam",
                "roundEndReason",
                "ctFreezeTimeEndEqVal",
                "ctRoundStartEqVal",
                "ctRoundSpendMoney",
                "ctBuyType",
                "tFreezeTimeEndEqVal",
                "tRoundStartEqVal",
                "tRoundSpendMoney",
                "tBuyType",
            ]
            for r in self.json["gameRounds"]:
                round_item = {}
                for k in cols:
                    round_item[k] = r[k]
                    round_item["matchID"] = self.json["matchID"]
                    round_item["mapName"] = self.json["mapName"]
                rounds.append(round_item)
            return pd.DataFrame(rounds)
        else:
            self.logger.error(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )
            raise AttributeError(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )

    def _parse_kills(self):
        """Returns kills as either a Pandas dataframe

        Returns:
            A Pandas dataframe where each row is a kill

        Raises:
            AttributeError: Raises an AttributeError if the .json attribute is None
        """
        if self.json:
            kills = []
            for r in self.json["gameRounds"]:
                if r["kills"] is not None:
                    for k in r["kills"]:
                        new_k = k
                        new_k["roundNum"] = r["roundNum"]
                        new_k["matchID"] = self.json["matchID"]
                        new_k["mapName"] = self.json["mapName"]
                        kills.append(new_k)
            return pd.DataFrame(kills)
        else:
            self.logger.error(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )
            raise AttributeError(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )

    def _parse_weapon_fires(self):
        """Returns weapon fires as either a list or Pandas dataframe

        Returns:
            A  Pandas dataframe where each row is a weapon fire event

        Raises:
            AttributeError: Raises an AttributeError if the .json attribute is None
        """
        if self.json:
            shots = []
            for r in self.json["gameRounds"]:
                if r["weaponFires"] is not None:
                    for wf in r["weaponFires"]:
                        new_wf = wf
                        new_wf["roundNum"] = r["roundNum"]
                        new_wf["matchID"] = self.json["matchID"]
                        new_wf["mapName"] = self.json["mapName"]
                        shots.append(new_wf)
            return pd.DataFrame(shots)
        else:
            self.logger.error(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )
            raise AttributeError(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )

    def _parse_damages(self):
        """Returns damages as a Pandas dataframe

        Returns:
            A Pandas dataframe where each row is a damage event.

        Raises:
            AttributeError: Raises an AttributeError if the .json attribute is None
        """
        if self.json:
            damages = []
            for r in self.json["gameRounds"]:
                if r["damages"] is not None:
                    for d in r["damages"]:
                        new_d = d
                        new_d["roundNum"] = r["roundNum"]
                        new_d["matchID"] = self.json["matchID"]
                        new_d["mapName"] = self.json["mapName"]
                        damages.append(new_d)
            return pd.DataFrame(damages)
        else:
            self.logger.error(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )
            raise AttributeError(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )

    def _parse_grenades(self):
        """Returns grenades as a Pandas dataframe

        Returns:
            A list or Pandas dataframe where each row is a grenade throw

        Raises:
            AttributeError: Raises an AttributeError if the .json attribute is None
        """
        if self.json:
            grenades = []
            for r in self.json["gameRounds"]:
                if r["grenades"] is not None:
                    for g in r["grenades"]:
                        new_g = g
                        new_g["roundNum"] = r["roundNum"]
                        new_g["matchID"] = self.json["matchID"]
                        new_g["mapName"] = self.json["mapName"]
                        grenades.append(new_g)
            return pd.DataFrame(grenades)
        else:
            self.logger.error(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )
            raise AttributeError(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )

    def _parse_bomb_events(self):
        """Returns bomb events as a Pandas dataframe

        Returns:
            A Pandas dataframe where each row is a bomb event (defuse, plant, etc.)

        Raises:
            AttributeError: Raises an AttributeError if the .json attribute is None
        """
        if self.json:
            bomb_events = []
            for r in self.json["gameRounds"]:
                if r["bombEvents"] is not None:
                    for b in r["bombEvents"]:
                        new_b = b
                        new_b["roundNum"] = r["roundNum"]
                        new_b["matchID"] = self.json["matchID"]
                        new_b["mapName"] = self.json["mapName"]
                        bomb_events.append(new_b)
            return pd.DataFrame(bomb_events)
        else:
            self.logger.error(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )
            raise AttributeError(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )

    def _parse_flashes(self):
        """Returns flashes as a Pandas dataframe

        Returns:
            A Pandas dataframe where each row is a flash event.

        Raises:
            AttributeError: Raises an AttributeError if the .json attribute is None
        """
        if self.json:
            flashes = []
            for r in self.json["gameRounds"]:
                if r["flashes"] is not None:
                    for f in r["flashes"]:
                        new_f = f
                        new_f["roundNum"] = r["roundNum"]
                        new_f["matchId"] = self.json["matchID"]
                        new_f["mapName"] = self.json["mapName"]
                        flashes.append(new_f)
            return pd.DataFrame(flashes)
        else:
            self.logger.error(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )
            raise AttributeError(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )

    def clean_rounds(
        self,
        remove_no_frames=True,
        remove_warmups=True,
        remove_knifes=True,
        remove_bad_timings=True,
        remove_excess_players=True,
        remove_excess_kills=True,
        remove_bad_endings=True,
        remove_bad_scoring=True,
        return_type="json",
        save_to_json=True,
    ):
        """Cleans a parsed demofile JSON.

        Args:
            remove_no_frames (bool, optional): Remove rounds where there are no frames. Default to True.
            remove_warmups (bool, optional): Remove warmup rounds. Defaults to True.
            remove_knifes (bool, optional): Remove knife rounds. Defaults to True.
            remove_bad_timings (bool, optional): Remove bad timings. Defaults to True.
            remove_excess_players (bool, optional): Remove rounds with more than 5 players. Defaults to True.
            remove_excess_kills (bool, optional): Remove rounds with more than 10 kills. Defaults to True.
            remove_bad_endings (bool, optional): Remove rounds with bad round end reasons. Defaults to True.
            remove_bad_scoring (bool, optional): Remove rounds where the scoring is off (like scores going below the previous round's). Defaults to False.
            return_type (str, optional): Return JSON or DataFrame. Defaults to "json".
            save_to_json (bool, optional): Whether to write the JSON to a file. Defaults to True.

        Raises:
            AttributeError: Raises an AttributeError if the .json attribute is None

        Returns:
            dict: A dictionary of the cleaned demo.
        """
        if self.json:
            if remove_no_frames:
                self.remove_rounds_with_no_frames()
            if remove_warmups:
                self.remove_warmups()
            if remove_knifes:
                self.remove_knife_rounds()
            if remove_bad_timings:
                self.remove_time_rounds()
            if remove_excess_players:
                self.remove_excess_players()
            if remove_excess_kills:
                self.remove_excess_kill_rounds()
            if remove_bad_endings:
                self.remove_end_round()
            if remove_bad_scoring:
                self.remove_bad_scoring()
            self.renumber_rounds()
            # self.rescore_rounds() -- Need to edit to take into account half switches
            if save_to_json:
                self.write_json()
            if return_type == "json":
                return self.json
            elif return_type == "df":
                demo_data = self.parse_json_to_df()
                self.logger.info("Returned cleaned dataframe output")
                return demo_data
        else:
            self.logger.error(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )
            raise AttributeError(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )

    def write_json(self):
        """Rewrite the JSON file"""
        with open(self.outpath + "/" + self.output_file, "w", encoding="utf8") as fp:
            json.dump(self.json, fp, indent=(1 if self.json_indentation else None))

    def renumber_rounds(self):
        """Renumbers the rounds.

        Raises:
            AttributeError: Raises an AttributeError if the .json attribute has a "gameRounds" key.
        """
        if self.json["gameRounds"]:
            for i, r in enumerate(self.json["gameRounds"]):
                self.json["gameRounds"][i]["roundNum"] = i + 1
        else:
            self.logger.error(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )
            raise AttributeError(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )

    def rescore_rounds(self):
        """Rescore the rounds based on round end reason.

        Raises:
            AttributeError: Raises an AttributeError if the .json attribute has a "gameRounds" key.
        """
        if self.json["gameRounds"]:
            for i, r in enumerate(self.json["gameRounds"]):
                if i == 0:
                    self.json["gameRounds"][i]["tScore"] = 0
                    self.json["gameRounds"][i]["ctScore"] = 0
                    if self.json["gameRounds"][i]["winningSide"] == "ct":
                        self.json["gameRounds"][i]["endCTScore"] = 1
                        self.json["gameRounds"][i]["endTScore"] = 0
                    if self.json["gameRounds"][i]["winningSide"] == "t":
                        self.json["gameRounds"][i]["endCTScore"] = 0
                        self.json["gameRounds"][i]["endTScore"] = 1
                elif i > 0:
                    self.json["gameRounds"][i]["tScore"] = self.json["gameRounds"][
                        i - 1
                    ]["endTScore"]
                    self.json["gameRounds"][i]["ctScore"] = self.json["gameRounds"][
                        i - 1
                    ]["endCTScore"]
                    if self.json["gameRounds"][i]["winningSide"] == "ct":
                        self.json["gameRounds"][i]["endCTScore"] = (
                            self.json["gameRounds"][i]["ctScore"] + 1
                        )
                        self.json["gameRounds"][i]["endTScore"] = self.json[
                            "gameRounds"
                        ][i]["tScore"]
                    if self.json["gameRounds"][i]["winningSide"] == "t":
                        self.json["gameRounds"][i]["endCTScore"] = self.json[
                            "gameRounds"
                        ][i]["ctScore"]
                        self.json["gameRounds"][i]["endTScore"] = (
                            self.json["gameRounds"][i]["tScore"] + 1
                        )
        else:
            self.logger.error(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )
            raise AttributeError(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )

    def remove_bad_scoring(self):
        """Removes rounds where the scoring is bad.

        We loop through the rounds:
        If the round ahead has equal or less score, we do not add the current round.
        If the round ahead has +1 score, we add the current round

        Raises:
            AttributeError: Raises an AttributeError if the .json attribute is None
        """
        if self.json:
            cleaned_rounds = []
            for i, r in enumerate(self.json["gameRounds"]):
                current_round_total = (
                    r["tScore"] + r["endTScore"] + r["ctScore"] + r["endCTScore"]
                )
                if i < len(self.json["gameRounds"]) - 1:
                    lookahead_round = self.json["gameRounds"][i + 1]
                    lookahead_round_total = (
                        lookahead_round["tScore"]
                        + lookahead_round["endTScore"]
                        + lookahead_round["ctScore"]
                        + lookahead_round["endCTScore"]
                    )
                    if lookahead_round_total > current_round_total:
                        cleaned_rounds.append(r)
                    elif (r["endTScore"] == 16) & (r["endCTScore"] <= 14):
                        cleaned_rounds.append(r)
                    elif (r["endCTScore"] == 16) & (r["endTScore"] <= 14):
                        cleaned_rounds.append(r)
                else:
                    lookback_round = self.json["gameRounds"][i - 1]
                    lookback_round_total = (
                        lookback_round["tScore"]
                        + lookback_round["endTScore"]
                        + lookback_round["ctScore"]
                        + lookback_round["endCTScore"]
                    )
                    if current_round_total > lookback_round_total:
                        cleaned_rounds.append(r)
            self.json["gameRounds"] = cleaned_rounds
        else:
            self.logger.error(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )
            raise AttributeError(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )

    def remove_rounds_with_no_frames(self):
        """Removes rounds with no frames

        Raises:
            AttributeError: Raises an AttributeError if the .json attribute is None
        """
        if self.json:
            if not self.parse_frames:
                self.logger.warning(
                    "parse_frames is set to False, must be true for remove_no_frames to work. Skipping remove_no_frames."
                )
            else:
                cleaned_rounds = []
                for r in self.json["gameRounds"]:
                    if len(r["frames"]) > 0:
                        cleaned_rounds.append(r)
                self.json["gameRounds"] = cleaned_rounds
        else:
            self.logger.error(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )
            raise AttributeError(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )

    def remove_excess_players(self):
        """Removes rounds where there are more than 5 players on a side.

        Raises:
            AttributeError: Raises an AttributeError if the .json attribute is None
        """
        if self.json:
            if not self.parse_frames:
                self.logger.warning(
                    "parse_frames is set to False, must be true for remove_excess_players to work. Skipping remove_excess_players."
                )
            else:
                cleaned_rounds = []
                # Remove rounds where the number of players is too large
                for r in self.json["gameRounds"]:
                    if len(r["frames"]) > 0:
                        f = r["frames"][0]
                        if f["ct"]["players"] == None:
                            if f["t"]["players"] == None:
                                pass
                            elif len(f["t"]["players"]) <= 5:
                                cleaned_rounds.append(r)
                        elif len(f["ct"]["players"]) <= 5:
                            if (f["t"]["players"] == None) or (
                                len(f["t"]["players"]) <= 5
                            ):
                                cleaned_rounds.append(r)
                self.json["gameRounds"] = cleaned_rounds
        else:
            self.logger.error(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )
            raise AttributeError(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )

    def remove_warmups(self):
        """Removes warmup rounds.

        Raises:
            AttributeError: Raises an AttributeError if the .json attribute is None
        """
        if self.json:
            cleaned_rounds = []
            # Remove warmups where the demo may have started recording in the middle of a warmup round
            if "warmupChanged" in self.json["matchPhases"]:
                if len(self.json["matchPhases"]["warmupChanged"]) > 1:
                    last_warmup_changed = self.json["matchPhases"]["warmupChanged"][1]
                    for r in self.json["gameRounds"]:
                        if (r["startTick"] > last_warmup_changed) and (
                            not r["isWarmup"]
                        ):
                            cleaned_rounds.append(r)
                        if r["startTick"] == last_warmup_changed:
                            cleaned_rounds.append(r)
                else:
                    for r in self.json["gameRounds"]:
                        if not r["isWarmup"]:
                            cleaned_rounds.append(r)
            self.json["gameRounds"] = cleaned_rounds
        else:
            self.logger.error(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )
            raise AttributeError(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )

    def remove_end_round(self, bad_endings=["Draw", "Unknown", ""]):
        """Removes rounds with bad end reason.

        Args:
            bad_endings (list, optional): List of bad round end reasons. Defaults to ["Draw", "Unknown", ""].

        Raises:
            AttributeError: Raises an AttributeError if the .json attribute is None
        """
        if self.json:
            cleaned_rounds = []
            for r in self.json["gameRounds"]:
                if r["roundEndReason"] not in bad_endings:
                    cleaned_rounds.append(r)
            self.json["gameRounds"] = cleaned_rounds
        else:
            self.logger.error(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )
            raise AttributeError(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )

    def remove_knife_rounds(self):
        """Removes knife rounds.

        Raises:
            AttributeError: Raises an AttributeError if the .json attribute is None
        """
        if self.json:
            cleaned_rounds = []
            for r in self.json["gameRounds"]:
                if not r["isWarmup"]:
                    total_kills = len(r["kills"])
                    total_knife_kills = 0
                    if total_kills > 0:
                        for k in r["kills"]:
                            if k["weapon"] == "Knife":
                                total_knife_kills += 1
                    if (total_knife_kills != total_kills) | (total_knife_kills == 0):
                        cleaned_rounds.append(r)
            self.json["gameRounds"] = cleaned_rounds
        else:
            self.logger.error(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )
            raise AttributeError(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )

    def remove_excess_kill_rounds(self):
        """Removes rounds with more than 10 kills.

        Raises:
            AttributeError: Raises an AttributeError if the .json attribute is None
        """
        if self.json:
            cleaned_rounds = []
            for r in self.json["gameRounds"]:
                if not r["isWarmup"]:
                    if len(r["kills"]) <= 10:
                        cleaned_rounds.append(r)
            self.json["gameRounds"] = cleaned_rounds
        else:
            self.logger.error(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )
            raise AttributeError(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )

    def remove_time_rounds(self):
        """Remove rounds with odd round timings.

        Raises:
            AttributeError: Raises an AttributeError if the .json attribute is None
        """
        if self.json:
            cleaned_rounds = []
            for r in self.json["gameRounds"]:
                if (
                    (r["startTick"] <= r["endTick"])
                    and (r["startTick"] <= r["endOfficialTick"])
                    and (r["startTick"] <= r["freezeTimeEndTick"])
                ):
                    cleaned_rounds.append(r)
            self.json["gameRounds"] = cleaned_rounds
        else:
            self.logger.error(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )
            raise AttributeError(
                "JSON not found. Run .parse() or .read_json() if JSON already exists"
            )



from setuptools import setup, find_packages

from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="awpy",
    version="1.1.6",
    packages=find_packages(),
    # Project uses reStructuredText, so ensure that the docutils get
    # installed or upgraded on the target machine
    install_requires=[
        "pandas>=0.25.3",
        "numpy>=1.18.1",
        "scipy>=1.4.1",
        "matplotlib>=3.1.2",
        "networkx>=2.6.3",
        "textdistance>=4.2.0",
        "imageio>=2.9.0",
        "tqdm>=4.55.2",
    ],
    package_data={
        # If any package contains *.txt or *.rst files, include them:
        "": [
            "*.go",
            "data/map/*.png",
            "data/map/*.json",
            "data/nav/*.txt",
            "data/nav/*.csv",
            "*.mod",
            "*.sum",
        ]
    },
    # metadata to display on PyPI
    author="Peter Xenopoulos",
    author_email="xenopoulos@nyu.edu",
    description="Counter-Strike: Global Offensive data parsing, analysis and visualization functions",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="esports sports-analytics csgo counter-strike",
    url="https://github.com/pnxenopoulos/awpy",
    project_urls={
        "Issues": "https://github.com/pnxenopoulos/awpy/issues",
        "Documentation": "https://awpy.readthedocs.io/en/latest/?badge=latest",
        "GitHub": "https://github.com/pnxenopoulos/awpy/",
    },
    classifiers=["License :: OSI Approved :: MIT License"],
)



import os
import shutil
import imageio

from tqdm import tqdm

import matplotlib.pyplot as plt

from awpy.data import MAP_DATA


def plot_map(map_name="de_dust2", map_type="original", dark=False):
    """Plots a blank map.

    Args:
        map_name (string): Map to search
        map_type (string): "original" or "simpleradar"
        dark (boolean): Only for use with map_type="simpleradar". Indicates if you want to use the SimpleRadar dark map type

    Returns:
        matplotlib fig and ax
    """
    if map_type == "original":
        map_bg = imageio.imread(
            os.path.join(os.path.dirname(__file__), "")
            + """../data/map/{0}.png""".format(map_name)
        )
    else:
        col = "light"
        if dark:
            col = "dark"
        map_bg = imageio.imread(
            os.path.join(os.path.dirname(__file__), "")
            + """../data/map/{0}_{1}.png""".format(map_name, col)
        )
    fig, ax = plt.subplots()
    ax.imshow(map_bg, zorder=0)
    return fig, ax


# Position function courtesy of PureSkill.gg
def position_transform(map_name, position, axis):
    """Transforms an X or Y coordinate.

    Args:
        map_name (string): Map to search
        position (float): X or Y coordinate
        axis (string): Either "x" or "y" (lowercase)

    Returns:
        float
    """
    start = MAP_DATA[map_name][axis]
    scale = MAP_DATA[map_name]["scale"]
    if axis == "x":
        pos = position - start
        pos /= scale
        return pos
    elif axis == "y":
        pos = start - position
        pos /= scale
        return pos
    else:
        return None


def plot_positions(
    positions=[],
    colors=[],
    markers=[],
    map_name="de_ancient",
    map_type="original",
    dark=False,
    apply_transformation=False,
):
    """Plots player positions

    Args:
        positions (list): List of lists of length 2 ([[x,y], ...])
        colors (list): List of colors for each player
        markers (list): List of marker types for each player
        map_name (string): Map to search
        map_type (string): "original" or "simpleradar"
        dark (boolean): Only for use with map_type="simpleradar". Indicates if you want to use the SimpleRadar dark map type
        apply_transformation (boolean): Indicates if you need to also use position_transform() for the X/Y coordinates

    Returns:
        matplotlib fig and ax
    """
    f, a = plot_map(map_name=map_name, map_type=map_type, dark=dark)
    for p, c, m in zip(positions, colors, markers):
        if apply_transformation:
            a.scatter(
                x=position_transform(map_name, p[0], "x"),
                y=position_transform(map_name, p[1], "y"),
                c=c,
                marker=m,
            )
        else:
            a.scatter(x=p[0], y=p[1], c=c, marker=m)
    a.get_xaxis().set_visible(False)
    a.get_yaxis().set_visible(False)
    return f, a


def plot_round(
    filename, frames, map_name="de_ancient", map_type="original", dark=False
):
    """Plots a round and saves as a .gif. CTs are blue, Ts are orange, and the bomb is an octagon. Only use untransformed coordinates.

    Args:
        filename (string): Filename to save the gif
        frames (list): List of frames from a parsed demo
        markers (list): List of marker types for each player
        map_name (string): Map to search
        map_type (string): "original" or "simpleradar"
        dark (boolean): Only for use with map_type="simpleradar". Indicates if you want to use the SimpleRadar dark map type

    Returns:
        matplotlib fig and ax, saves .gif
    """
    if os.path.isdir("csgo_tmp"):
        shutil.rmtree("csgo_tmp/")
    os.mkdir("csgo_tmp")
    image_files = []
    for i, f in tqdm(enumerate(frames)):
        positions = []
        colors = []
        markers = []
        # Plot bomb
        # Thanks to https://github.com/pablonieto0981 for adding this code!
        for object_type in f["world"]:
            if object_type["objectType"] == "bomb":
                colors.append("orange")
                markers.append("8")
                pos = (
                    position_transform(map_name, object_type["x"], "x"),
                    position_transform(map_name, object_type["y"], "y"),
                )
                positions.append(pos)
            else:
                pass
        # Plot players
        for side in ["ct", "t"]:
            for p in f[side]["players"]:
                if side == "ct":
                    colors.append("cyan")
                else:
                    colors.append("red")
                if p["hp"] == 0:
                    markers.append("x")
                else:
                    markers.append(".")
                pos = (
                    position_transform(map_name, p["x"], "x"),
                    position_transform(map_name, p["y"], "y"),
                )
                positions.append(pos)
        f, a = plot_positions(
            positions=positions,
            colors=colors,
            markers=markers,
            map_name=map_name,
            map_type=map_type,
            dark=dark,
        )
        image_files.append("csgo_tmp/{}.png".format(i))
        f.savefig(image_files[-1], dpi=300, bbox_inches="tight")
        plt.close()
    images = []
    for file in image_files:
        images.append(imageio.imread(file))
    imageio.mimsave(filename, images)
    shutil.rmtree("csgo_tmp/")
    return True


def plot_nades(
    rounds, nades=[], side="CT", map_name="de_ancient", map_type="original", dark=False
):
    """Plots grenade trajectories.

    Args:
        rounds (list): List of round objects from a parsed demo
        nades (list): List of grenade types to plot
        side (string): Specify side to plot grenades. Either "CT" or "T".
        map_name (string): Map to search
        map_type (string): "original" or "simpleradar"
        dark (boolean): Only for use with map_type="simpleradar". Indicates if you want to use the SimpleRadar dark map type

    Returns:
        matplotlib fig and ax
    """
    f, a = plot_map(map_name=map_name, map_type=map_type, dark=dark)
    for r in rounds:
        if r["grenades"]:
            for g in r["grenades"]:
                if g["throwerSide"] == side:
                    start_x = position_transform(map_name, g["throwerX"], "x")
                    start_y = position_transform(map_name, g["throwerY"], "y")
                    end_x = position_transform(map_name, g["grenadeX"], "x")
                    end_y = position_transform(map_name, g["grenadeY"], "y")
                    if g["grenadeType"] in nades:
                        if (
                            g["grenadeType"] == "Incendiary Grenade"
                            or g["grenadeType"] == "Molotov"
                        ):
                            a.plot([start_x, end_x], [start_y, end_y], color="red")
                            a.scatter(end_x, end_y, color="red")
                        if g["grenadeType"] == "Smoke Grenade":
                            a.plot([start_x, end_x], [start_y, end_y], color="gray")
                            a.scatter(end_x, end_y, color="gray")
                        if g["grenadeType"] == "HE Grenade":
                            a.plot([start_x, end_x], [start_y, end_y], color="green")
                            a.scatter(end_x, end_y, color="green")
                        if g["grenadeType"] == "Flashbang":
                            a.plot([start_x, end_x], [start_y, end_y], color="gold")
                            a.scatter(end_x, end_y, color="gold")
    a.get_xaxis().set_visible(False)
    a.get_yaxis().set_visible(False)
    return f, a
